﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml;
using System.Threading;
using System.Net;

using eeSignAnyWhere_SampleCode_SOAP;

namespace eSignAnyWhere_SampleCode_SOAP
{
    class Program
    {

        public const string eSawUrl = "https://www.significant.com";
        const string UserAccountName = "your@email.com";
        const string OrganizationKey = "48024f64-c2b4-403a-ac16-80c287b2fa2d"; //they value in the eSAW UI: Settings/Organization/OrganizationDetails -> Organization Key

        const string TemplateEnvelopeID = "4c1ba196-833b-4626-af0a-a387c6d6a4a6"; 



        /* Please note, that eSignAnywhere consist in reality of 2 components: 
         *                a.) The workflow management system eSignAnywhere  - and - 
         *                b.) The system that takes care about each single step in the workflow: Workstep Controller and SAW Viewer (the front end for signing the document). Both are part of SIGNificant Server Platform (SSP)
         *                
         * The sample code shown below focuses on the eSignAnywhere specific part of the system. 
         * Whenever if comes to a specific configuration of a single step in the workflow, or for a dedicated behavior of Saw Viewer, that you want to control, please have a look here as well:
         *       
         *      https://partner.xyzmo.com/projects/significant-server-platform/wiki/WSC_HowTo_Scenarios
         *      https://partner.xyzmo.com/projects/significant-server-platform/wiki/WSC_HowTo_Scenarios#Full-WorkstepConfiguration
         *      https://partner.xyzmo.com/projects/significant-server-platform/wiki/SAW_Integration_Guide#ViewerPreferences
         * 
         * In case you can't access these resources, please get into contact with your presales contact or contact our support.
         * 
         */

        static void Main(string[] args)
        {
            ValidateAuthorization();

            //SendEnvelope_Simple(new string[] { @"..\..\..\Ressources\Demo_Contract_Form.pdf" });//,  @"..\..\..\Ressources\Title.pdf"});
            //SendEnvelope_with2Recipients(false);
            //SendEnvelope_with2Recipients(true); //bulk envelope
            //SendEnvelope_Simple_AddAdditionalSignatureFields();
            //SendEnvelope_Simple_PrefillFormFieldValues(new string[] { @"..\..\..\Ressources\Demo_Contract_Form.pdf" });
            //SendEnvelope_fromDocumentWithTags();
            //SendEnvelope_Simple_DocumentTimestamp_And_AutomaticSignature(new string[] { @"..\..\..\Ressources\Demo_Contract_Form.pdf" });
            //SendEnvelope_Simple_RemoteSignatureAndSmsOtpSignature_withPredefinedSettingsForSigner(new string[] { @"..\..\..\Ressources\Demo_Contract_Form.pdf" });
            //SendEnvelope_Simple_DisposableRemoteCertificate(new string[] { @"..\..\..\Ressources\Demo_Contract_Form.pdf" });
            //SendEnvelope_inPosScenario();

            //CreateDraft(new string[] { @"..\..\..\Ressources\Demo_Contract_Form.pdf", @"..\..\..\Ressources\Title.pdf" });            

            //SendEnvelopeFromTemplate(TemplateEnvelopeID);
            //UsingATemplateForSendEnvelope(TemplateEnvelopeID);
            //CreateDraftFromTemplate(TemplateEnvelopeID);

            //DownloadCompletedEnvelope("4cc54e65-1484-4e59-b3d0-122e62372763");

            //SendReminder("e6c66e34-61e1-4fbf-8459-e45b4d95a48e");
            //CancelEnvelope("your envelopeID");
            //RestartEnvelope("your envelopeID");
            //DeleteEnvelope("your envelopeID");

            GetEnvelopeStatus("88a8d99c-f2b9-45c5-aa8f-fdc46b0a98c1");
            //FindEnvelops();
            //FindAllEnvelopsThatAreWaitingForDedicatedRecipientAndReplaceRecipient();

            //ReplaceOrChangeRecipientInEnvelope();
            //ChangeRecipientsTasksInEnvelope();
            //DeleteRecipientFromEnvelope();



            // --- Functions for user managment ---
                   
            //UserManagement.FindUsers();
            //UserManagement.CreateUser();
            //UserManagement.UpdateUser();
            //UserManagement.DeleteUser();
            //UserManagement.GetTeamsAndSetTeams();

            Console.ReadLine();
        }

        #region Helper Methods

        private static eSAW.Api GetSawWebService()
        {
            eSAW.Api ws = new eSAW.Api();
            ws.Url = eSawUrl + @"/api.asmx";

            // for onPremise version with WindowsAuthentication only: set user token
            ws.Credentials = System.Net.CredentialCache.DefaultCredentials;

            return ws;
        }

        public static string GetApiAuthorization()
        {
            string authorisationXml = "<authorization><organizationKey>##OrgKey##</organizationKey><userLoginName>##UserLoginName##</userLoginName></authorization>";

            authorisationXml = authorisationXml.Replace("##OrgKey##", OrganizationKey);
            authorisationXml = authorisationXml.Replace("##UserLoginName##", UserAccountName);

            return authorisationXml;
        }

        private static string[] UploadPdfsAndGetDocIds(eSAW.Api ws, string authXml, string[] documentPaths)
        {
            string[] docIds = new string[documentPaths.Length];
            for (int i = 0; i < docIds.Length; i++)
            {
                string docXml = GetDocXml(documentPaths[i]);
                string fileUploadResponseXml = ws.UploadTemporarySspFile_v1(authXml, docXml);
                SignAnywhereResponse uploadResult = SignAnywhereResponse.parseResponse(fileUploadResponseXml);
                if (uploadResult.BaseResult == SignAnywhereResponse.Result.Ok)
                {
                    docIds[i] = uploadResult.okInfo.InnerText;
                    Console.WriteLine("Document upload successful! DocId: " + uploadResult.okInfo.InnerText);
                }
                else
                {
                    Console.WriteLine(uploadResult.Error + Environment.NewLine + uploadResult.ErrorMsg);
                    throw new ApplicationException(uploadResult.ErrorMsg);
                }


            }
            return docIds;
        }

        private static string GetDocXml(string filePath)
        {
            SignAnywhereFile sawFile = new SignAnywhereFile();
            FileInfo pdfFile = new FileInfo(filePath);
            sawFile.Data= File.ReadAllBytes(pdfFile.FullName);
            sawFile.FileName = pdfFile.Name;

            return sawFile.ToXml();                   
        }
        #endregion

        #region Sending Envelops

        public static void SendEnvelope_Simple( string [] documentsPath)
        {
            try
            {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();

                string[] docIds = UploadPdfsAndGetDocIds(ws, auth, documentsPath);

                #region Analyze Pdf from Server
                // this can be ommitted, if you have already a workstep configuration that fits to your document(s)

                string adHocConfig = File.ReadAllText(@"..\..\..\Ressources\AdHocConfig.xml");                
                string getAdHocWSConfigResponseXml = ws.GetAdHocWorkstepConfiguration_v1(auth, docIds, adHocConfig);

                SignAnywhereResponse getAdHocWsConfigResponse = SignAnywhereResponse.parseResponse(getAdHocWSConfigResponseXml);
                if (getAdHocWsConfigResponse.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(getAdHocWsConfigResponse.Error + Environment.NewLine + getAdHocWsConfigResponse.ErrorMsg);
                    return;
                }

                // note for on premise versions: make sure to have the correct signing certificate configured in your workstep configuration.
                //                               workstepConfiguration/signaturePluginConfiguration/PdfSignatureCryptographicData_V1/SigningCertificateDescriptor/Identifier
                #endregion
                
                //
                string envelopeConfig = File.ReadAllText(@"..\..\..\Ressources\EnvelopeConfigTemplate.xml");        

                //envelopeConfig = envelopeConfig.Replace("<!--##WorkstepConfig##-->", getAdHocWsConfigResponse.okInfo.InnerXml);

                //XmlDocument wscConfig = new XmlDocument();
                //wscConfig.LoadXml(getAdHocWsConfigResponse.okInfo.InnerXml);
                //XmlElement newElem = wscConfig.CreateElement("MyCustomElement");
                //newElem.InnerText = "My Sample Value to investigate";
                //wscConfig.DocumentElement.AppendChild(newElem);
                //envelopeConfig = envelopeConfig.Replace("<!--##WorkstepConfig##-->", wscConfig.OuterXml);

                envelopeConfig = envelopeConfig.Replace("<!--##WorkstepConfig##-->", File.ReadAllText(@"E:\Temp\wsc_customDisclaimer.config.xml"));

                string sendEnvelopeResponseXml = ws.SendEnvelope_v1(auth, docIds, envelopeConfig);
                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(sendEnvelopeResponseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg, "Error");
                    return;
                }
                Console.WriteLine("EnvelopeÍD: " + response.okInfo.InnerText);

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }

        }


        // this sample shows, how to send an envelope to 2 recipients and how to assign fields (signatures, form fields) to each recipient
        // new with 3.0: the EnvelopeConfigTemplate_2Recipients.xml shows how to hide a document for a certain recipient
        // new with 3.1: the EnvelopeConfigTemplate_2RecipientsBulk.xml shows, how to send a bulk transaction. This feature needs to be enabled in your subscription. Contact our sales for this.
        public static void SendEnvelope_with2Recipients(bool bulkEnvelope)
        {
            try
            {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();

                string[] documentsPath = new string[] { @"..\..\..\Ressources\Demo_Contract_Form.pdf", @"..\..\..\Ressources\Title.pdf" };
                string[] docIds = UploadPdfsAndGetDocIds(ws, auth, documentsPath);

                #region Analyze Pdf from Server and adjust per recipient
                // this can be ommitted, if you have already a workstep configuration that fits to your document(s) and recipients
                // this code shows, how to programmatically adjust the configuration xml, however, this can be done manually as well and loaded as config

                string adHocConfig = File.ReadAllText(@"..\..\..\Ressources\AdHocConfig.xml");
                string getAdHocWSConfigResponseXml = ws.GetAdHocWorkstepConfiguration_v1(auth, docIds, adHocConfig);

                SignAnywhereResponse getAdHocWsConfigResponse = SignAnywhereResponse.parseResponse(getAdHocWSConfigResponseXml);
                if (getAdHocWsConfigResponse.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(getAdHocWsConfigResponse.Error + Environment.NewLine + getAdHocWsConfigResponse.ErrorMsg);
                    return;
                }               
                
                // signature13 and Dropdown_SalesAgentName are for second recipient, all other fields and signatures are for first recipient
                // therefore adjust the xml accordingly
                // in addition, the sample demonstrate how to set a task as required
                #region  adjust workstep config for recipient 1               
                
                string workstepConfigRecipient1 = getAdHocWsConfigResponse.okInfo.InnerXml;
                XmlDocument wsConfigXmlRecipient1 =  new XmlDocument();
                wsConfigXmlRecipient1.LoadXml(workstepConfigRecipient1);                

                XmlNode signatureNodeRecip1 = wsConfigXmlRecipient1.DocumentElement.SelectSingleNode("signatureTemplate");
                XmlNode workstepTasksNodeRecip1 = wsConfigXmlRecipient1.DocumentElement.SelectSingleNode("Policy/WorkstepTasks");
                XmlNode formElementsNodeRecip1 = wsConfigXmlRecipient1.DocumentElement.SelectSingleNode("pdfForms");
                
                // set signature12 to required (in the task list and in the detail for the signature
                XmlNode sig12DetailNode = signatureNodeRecip1.SelectSingleNode("sig[@id=\"Signature12\"]");
                sig12DetailNode.SelectSingleNode("param[@name = \"req\"]").InnerText = "1";

                XmlNode sig12Task = workstepTasksNodeRecip1.SelectSingleNode("Task[@id=\"Signature12\"]");
                ((XmlAttribute)sig12Task.SelectSingleNode("@required")).Value = "1";
                ((XmlAttribute)sig12Task.SelectSingleNode("@displayName")).Value = "My custom text for click to sign"; //set to "" to get the default text in the language choosen by the user.
                
                // remove signature13
                signatureNodeRecip1.RemoveChild(signatureNodeRecip1.SelectSingleNode("sig[@id=\"Signature13\"]"));
                workstepTasksNodeRecip1.RemoveChild(workstepTasksNodeRecip1.SelectSingleNode("Task[@id=\"Signature13\"]"));

                // remove formfield Dropdown_SalesAgentName
                XmlNode salesAgentRecip1 = formElementsNodeRecip1.SelectSingleNode("pdfFormsGroup/*[@id=\"Dropdown_SalesAgentName\"]");
                salesAgentRecip1.ParentNode.RemoveChild(salesAgentRecip1);
                //
                workstepConfigRecipient1 = wsConfigXmlRecipient1.OuterXml;
                #endregion // recipient 1

                #region adjust workstep config for recipient 2
                string workstepConfigRecipient2 = getAdHocWsConfigResponse.okInfo.InnerXml;

                XmlDocument wsConfigXmlRecipient2 = new XmlDocument();
                wsConfigXmlRecipient2.LoadXml(workstepConfigRecipient2);

                XmlNode signatureNodeRecip2 = wsConfigXmlRecipient2.DocumentElement.SelectSingleNode("signatureTemplate");
                XmlNode workstepTasksNodeRecip2 = wsConfigXmlRecipient2.DocumentElement.SelectSingleNode("Policy/WorkstepTasks");
                XmlNode formElementsNodeRecip2 = wsConfigXmlRecipient2.DocumentElement.SelectSingleNode("pdfForms");

                // set signature13 to required (in the task list and in the detail for the signature
                XmlNode sig13DetailNode = signatureNodeRecip2.SelectSingleNode("sig[@id=\"Signature13\"]");
                sig13DetailNode.SelectSingleNode("param[@name = \"req\"]").InnerText = "1";

                XmlNode sig13Task = workstepTasksNodeRecip2.SelectSingleNode("Task[@id=\"Signature13\"]");
                ((XmlAttribute)sig13Task.SelectSingleNode("@required")).Value = "1";

                // set formfield Dropdown_SalesAgentName as reqired
                XmlNode salesAgentRecip2 = formElementsNodeRecip2.SelectSingleNode("pdfFormsGroup/*[@id=\"Dropdown_SalesAgentName\"]");
                XmlNode salesAgentRecip2Req = salesAgentRecip2.SelectSingleNode("isRequired");
                if (salesAgentRecip2Req != null)
                {
                    salesAgentRecip2Req.InnerText = "1";
                }
                else {
                    salesAgentRecip2Req= wsConfigXmlRecipient2.CreateElement("isRequired");
                    salesAgentRecip2Req.InnerText = "1";
                    salesAgentRecip2.AppendChild(salesAgentRecip2Req);
                }

                XmlNode formFillingTaskRecip2 = workstepTasksNodeRecip2.SelectSingleNode("Task[@type=\"FillFormsGroup\"]");
                ((XmlAttribute)formFillingTaskRecip2.SelectSingleNode("@required")).Value = "1";

                // remove signature12
                signatureNodeRecip2.RemoveChild(signatureNodeRecip2.SelectSingleNode("sig[@id=\"Signature12\"]"));
                workstepTasksNodeRecip2.RemoveChild(workstepTasksNodeRecip2.SelectSingleNode("Task[@id=\"Signature12\"]"));

                // remove all other form fields
                XmlNode formFillGroupNode = salesAgentRecip2.ParentNode;
                XmlAttribute formGroupId = ((XmlElement)formFillGroupNode).GetAttributeNode("id");
                formFillGroupNode.RemoveAll();
                ((XmlElement)formFillGroupNode).SetAttributeNode(formGroupId);
                formFillGroupNode.AppendChild(salesAgentRecip2);

                //
                workstepConfigRecipient2 = wsConfigXmlRecipient2.OuterXml;

                #endregion //recipient 2
                //
                #endregion
                //
                string envelopeConfig = "";
                if (bulkEnvelope)
                {
                    // 2nd step is bulk step ... just define additional recipients in the recipients section of step 2
                    envelopeConfig = File.ReadAllText(@"..\..\..\Ressources\EnvelopeConfigTemplate_2RecipientsBulk.xml");

                }
                else
                {
                    envelopeConfig = File.ReadAllText(@"..\..\..\Ressources\EnvelopeConfigTemplate_2Recipients.xml");
                }
                
                envelopeConfig = envelopeConfig.Replace("<!--##WorkstepConfigRecipient1##-->", workstepConfigRecipient1);
                envelopeConfig = envelopeConfig.Replace("<!--##WorkstepConfigRecipient2##-->", workstepConfigRecipient2);
                //

                string sendEnvelopeResponseXml = ws.SendEnvelope_v1(auth, docIds, envelopeConfig);
                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(sendEnvelopeResponseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg, "Error");
                    return;
                }

                if (bulkEnvelope)
                {
                    /*
                    <bulk id="dea27ba5-84ef-4435-96e4-3f353e5256c6">
	                    <envelopeId eMail="bulkRecip1@mail.com">20f4ee12-48e9-45e0-bfd5-884aa5bb42bd</envelopeId>
	                    <envelopeId eMail="bulkRecip2@mail.com">dd96e2f0-a373-470f-82d2-f08885167fa1</envelopeId>
                    </bulk>
                     */

                    XmlElement bulkElem = (XmlElement)response.okInfo.SelectSingleNode("bulk");
                    Console.WriteLine("Bulk EnvelopeId: " + bulkElem.GetAttribute("id"));
                    
                    XmlNodeList subEnvelopes = bulkElem.SelectNodes("envelopeId");
                    for (int i = 0; i < subEnvelopes.Count; i++)
                    {
                        Console.WriteLine("\tID: "+ subEnvelopes[i].InnerText +"\t email: " + ((XmlElement)subEnvelopes[i]).GetAttribute("eMail"));
                    }
                }
                else
                {
                    Console.WriteLine("EnvelopeÍD: " + response.okInfo.InnerText);
                }

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }

        }


        //this sample shows, how to add a new signature field and some new annotations for a recipient
        public static void SendEnvelope_Simple_AddAdditionalSignatureFields()
        {
            try
            {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();

                string[] documentsPath = new string[] { @"..\..\..\Ressources\Title.pdf" };
                string[] docIds = UploadPdfsAndGetDocIds(ws, auth, documentsPath);

                #region Analyze Pdf from Server and adjust per recipient
                // this can be ommitted, if you have already a workstep configuration that fits to your document(s) and recipients
                // this code shows, how to programmatically adjust the configuration xml, however, this can be done manually as well and loaded as config

                string adHocConfig = File.ReadAllText(@"..\..\..\Ressources\AdHocConfig.xml");
                string getAdHocWSConfigResponseXml = ws.GetAdHocWorkstepConfiguration_v1(auth, docIds, adHocConfig);

                SignAnywhereResponse getAdHocWsConfigResponse = SignAnywhereResponse.parseResponse(getAdHocWSConfigResponseXml);
                if (getAdHocWsConfigResponse.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(getAdHocWsConfigResponse.Error + Environment.NewLine + getAdHocWsConfigResponse.ErrorMsg);
                    return;
                }

                // place an additional signature field and additional prefilled annotations for name of signer and for date
                // therefore adjust the xml accordingly
              
                #region  adjust workstep config for recipient

                string workstepConfigRecipient1 = getAdHocWsConfigResponse.okInfo.InnerXml;
                XmlDocument wsConfigXmlRecipient1 = new XmlDocument();
                wsConfigXmlRecipient1.LoadXml(workstepConfigRecipient1);

                XmlNode signatureNodeRecip1 = wsConfigXmlRecipient1.DocumentElement.SelectSingleNode("signatureTemplate");
                XmlNode workstepTasksNodeRecip1 = wsConfigXmlRecipient1.DocumentElement.SelectSingleNode("Policy/WorkstepTasks");

                // add annotation task  
                string typewriterAnnottxt = File.ReadAllText(@"..\..\..\Ressources\TypewriterAnnotationTaskInfo.xml");
                XmlDocument typewriterAnnot = new XmlDocument();
                typewriterAnnot.LoadXml(typewriterAnnottxt);
                XmlNode typeWriterAnnotNode = wsConfigXmlRecipient1.ImportNode(typewriterAnnot.DocumentElement, true);
                wsConfigXmlRecipient1.DocumentElement.AppendChild(typeWriterAnnotNode);

                XmlNodeList typeWriterNodesList = typeWriterAnnotNode.SelectNodes("typewriterAnnotation");
                for (int t = 0; t < typeWriterNodesList.Count; t++)
                {
                   XmlElement task = wsConfigXmlRecipient1.CreateElement("Task");
                   task.SetAttribute("id",((XmlElement)typeWriterNodesList[t]).GetAttribute("id"));
                   task.SetAttribute("type", "AddTypewriterAnnotation");
                   task.SetAttribute("displayName", "");
                   task.SetAttribute("DocRefNumber", ((XmlElement)typeWriterNodesList[t]).SelectSingleNode("DocRefNumber").InnerText);

                   workstepTasksNodeRecip1.AppendChild(task);
                }
                
                // add sigfield
                string sigFieldTxt = File.ReadAllText(@"..\..\..\Ressources\SigFieldDefinition.xml");
                XmlDocument sigFieldDoc = new XmlDocument();
                sigFieldDoc.LoadXml(sigFieldTxt);
                XmlNode sfNodeNode = wsConfigXmlRecipient1.ImportNode(sigFieldDoc.DocumentElement, true);
                signatureNodeRecip1.AppendChild(sfNodeNode);

                XmlElement sftask = wsConfigXmlRecipient1.CreateElement("Task");
                sftask.SetAttribute("id", ((XmlElement)sfNodeNode).GetAttribute("id"));
                sftask.SetAttribute("type", "SignField");
                sftask.SetAttribute("DocRefNumber", ((XmlElement)sfNodeNode).SelectSingleNode("DocRefNumber").InnerText);
                sftask.SetAttribute("required", "1");
                sftask.SetAttribute("displayName", "Your Label for Click2Sign");
                workstepTasksNodeRecip1.AppendChild(sftask);
                
                //
                workstepConfigRecipient1 = wsConfigXmlRecipient1.OuterXml;
                #endregion // recipient 1

                //
                #endregion
                //
                string envelopeConfig = File.ReadAllText(@"..\..\..\Ressources\EnvelopeConfigTemplate_wAuth.xml");
                envelopeConfig = envelopeConfig.Replace("<!--##WorkstepConfig##-->", workstepConfigRecipient1);
                //

                string sendEnvelopeResponseXml = ws.SendEnvelope_v1(auth, docIds, envelopeConfig);
                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(sendEnvelopeResponseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg, "Error");
                    return;
                }
                Console.WriteLine("EnvelopeÍD: " + response.okInfo.InnerText);

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }

        }

        // this sample shows, how to prefill form field values before sending an envelope
        // obviously, the form field values in the pdf file(s) could be changed already prior of uploading the pdf file to eSignAnywhere, 
        // but in case this does not fit for your environment, eSignAnywhere allows you to do so as well
        //
        // 1) add <overrideFormFieldValues> to envelope configuration
        public static void SendEnvelope_Simple_PrefillFormFieldValues(string[] documentsPath)
        {
            try
            {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();

                string[] docIds = UploadPdfsAndGetDocIds(ws, auth, documentsPath);

                #region Analyze Pdf from Server and adjust per recipient
                // this can be ommitted, if you have already a workstep configuration that fits to your document(s) and recipients
                // this code shows, how to programmatically adjust the configuration xml, however, this can be done manually as well and loaded as config

                string adHocConfig = File.ReadAllText(@"..\..\..\Ressources\AdHocConfig.xml");
                string getAdHocWSConfigResponseXml = ws.GetAdHocWorkstepConfiguration_v1(auth, docIds, adHocConfig);

                SignAnywhereResponse getAdHocWsConfigResponse = SignAnywhereResponse.parseResponse(getAdHocWSConfigResponseXml);
                if (getAdHocWsConfigResponse.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(getAdHocWsConfigResponse.Error + Environment.NewLine + getAdHocWsConfigResponse.ErrorMsg);
                    return;
                }
                #endregion

                //
                string envelopeConfig = File.ReadAllText(@"..\..\..\Ressources\EnvelopeConfigTemplate_wAuth.xml");
                envelopeConfig = envelopeConfig.Replace("<!--##WorkstepConfig##-->", getAdHocWsConfigResponse.okInfo.InnerXml);

                /* Override form field values specification
                 * Can be used with this web services: SendEnvelope_v1, SendEnvelopeFromTemplate_v1, CreateDraft_v1 and CreateDraftFromTemplate_v1
                 
                   <envelope> 
                    ...
                       <!-- overrideOptions -->
                       <overrideFormFieldValues>
                          <document docRef="1">
                             <textBox name="formFieldId1">
                                <value>textBoxValue</value>
                             </textBox>
                             <listBox name="formFieldId2">
                                <selectedItems>
                                   <selectedItemId>selectedItem1</selectedItemId>
                                   <selectedItemId>selectedItem2</selectedItemId>
                                </selectedItems>
                             </listBox>
                             <radioButtonGroup name="formFieldId3">
                                <selectedItemId>selectedItem</selectedItemId>
                             </radioButtonGroup>
                             <checkBox name="formFieldId4">
                                <isChecked>true|false|0|1</isChecked>
                             </checkBox>
                             <comboBox name="formFieldId5">
                                <value>comboBoxValue</value>
                             </comboBox>
                          </document>
                       </overrideFormFieldValues>
                    ...
                    </envelope> 
                */
                envelopeConfig = envelopeConfig.Replace("<!--##OverrideFormFieldValues##-->", File.ReadAllText(@"..\..\..\Ressources\Demo_Contract_Form_OverrideFormFieldValues.xml"));

                string sendEnvelopeResponseXml = ws.SendEnvelope_v1(auth, docIds, envelopeConfig);
                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(sendEnvelopeResponseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg, "Error");
                    return;
                }
                Console.WriteLine("EnvelopeÍD: " + response.okInfo.InnerText);

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }

        }

        // this sample shows, how to work with a pdf file with tags for form fields and signature fields inside.
        // differences to the simple sendEnvelope scenario:
        // 1) use PrepareSendEnvelopeSteps_v1 to analyze the document, instead of the (optional) GetAdHocWorkstepConfiguration_v1
        // 2) replace in the envelope configuration the placeholder for the recipients by the real data for the recipients
        // 3) (optional): modify data in form fields parsed out of the tags

        public static void SendEnvelope_fromDocumentWithTags()
        {
            try
            {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();

                //string[] documentsPath = new string[] { @"..\..\..\Ressources\FieldMarkUpSample.pdf" };
                string[] documentsPath = new string[] { @"..\..\..\Ressources\Demo_Contract_Form.pdf" };
                string[] docIds = UploadPdfsAndGetDocIds(ws, auth, documentsPath);

                //1) use PrepareSendEnvelopeSteps_v1 to analyze the document, instead of the (optional) GetAdHocWorkstepConfiguration_v1
                string adHocConfig = File.ReadAllText(@"..\..\..\Ressources\AdHocConfig.xml");
                string prepareSendEnvelopeConfig = "<prepareSendEnvelopeStepsDescriptor>  <clearFieldMarkupString>1</clearFieldMarkupString></prepareSendEnvelopeStepsDescriptor>";
                string prepareSendEnvelopeStepsResponseXml = ws.PrepareSendEnvelopeSteps_v1(auth, docIds, adHocConfig, prepareSendEnvelopeConfig);
                SignAnywhereResponse prepareSendEnvelopeStepsResponse = SignAnywhereResponse.parseResponse(prepareSendEnvelopeStepsResponseXml);
                if (prepareSendEnvelopeStepsResponse.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(prepareSendEnvelopeStepsResponse.Error + Environment.NewLine + prepareSendEnvelopeStepsResponse.ErrorMsg);
                    return;
                }

                string envelopeConfigStr = File.ReadAllText(@"..\..\..\Ressources\EnvelopeConfigTemplate_wAuth.xml");
                XmlDocument envelopeConfig = new XmlDocument();
                envelopeConfig.LoadXml(envelopeConfigStr);

                envelopeConfig.DocumentElement.RemoveChild(envelopeConfig.DocumentElement.SelectSingleNode("steps"));
                // add the node <steps> and <addFormFields> into the envelope config
                envelopeConfig.DocumentElement.AppendChild(envelopeConfig.ImportNode(prepareSendEnvelopeStepsResponse.okInfo.SelectSingleNode("prepareSendEnvelopeStepsResult/steps"), true));
                envelopeConfig.DocumentElement.AppendChild(envelopeConfig.ImportNode(prepareSendEnvelopeStepsResponse.okInfo.SelectSingleNode("prepareSendEnvelopeStepsResult/addFormFields"), true));

                // 2) replace in the envelope configuration the placeholder for the recipients by the real data for the recipients
                // adopt the recipient settings for each step
                XmlNodeList recipientsNodeList = envelopeConfig.DocumentElement.SelectNodes("steps/step/recipients/recipient");
                recipientsNodeList[0].SelectSingleNode("eMail").InnerText = "your@email.com";
                recipientsNodeList[0].SelectSingleNode("firstName").InnerText = "first";
                recipientsNodeList[0].SelectSingleNode("lastName").InnerText = "signername";

                recipientsNodeList[1].SelectSingleNode("eMail").InnerText = "your.second@email.com";
                recipientsNodeList[1].SelectSingleNode("firstName").InnerText = "second";
                recipientsNodeList[1].SelectSingleNode("lastName").InnerText = "signername";

                // 3) (optional): modify data in form fields parsed out of the tags 
                XmlNode formFieldNameSigner2Node = envelopeConfig.DocumentElement.SelectSingleNode("addFormFields/document/textBox[@name='nameSig2']");
                formFieldNameSigner2Node.SelectSingleNode("value").InnerText = "second signername";

                // send the envelope
                string sendEnvelopeResponseXml = ws.SendEnvelope_v1(auth, docIds,envelopeConfig.OuterXml);
                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(sendEnvelopeResponseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg, "Error");
                    return;
                }
                Console.WriteLine("EnvelopeÍD: " + response.okInfo.InnerText);

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }

        }

        // this sample shows, how to declare for a recipient, that all documents should get a document timestamp after the recipient clicked on Finish.
        // to be able to use this feature, you need this feature enabled. contact our sales team for this.
        //
        // 1) in the workstep config for the recipient, add document timestamp
        public static void SendEnvelope_Simple_DocumentTimestamp_And_AutomaticSignature(string[] documentsPath)
        {
            try
            {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();

                string[] docIds = UploadPdfsAndGetDocIds(ws, auth, documentsPath);

                #region Analyze Pdf from Server and adjust per recipient
                // this can be ommitted, if you have already a workstep configuration that fits to your document(s) and recipients
                // this code shows, how to programmatically adjust the configuration xml, however, this can be done manually as well and loaded as config

                string adHocConfig = File.ReadAllText(@"..\..\..\Ressources\AdHocConfig.xml");
                string getAdHocWSConfigResponseXml = ws.GetAdHocWorkstepConfiguration_v1(auth, docIds, adHocConfig);

                SignAnywhereResponse getAdHocWsConfigResponse = SignAnywhereResponse.parseResponse(getAdHocWSConfigResponseXml);
                if (getAdHocWsConfigResponse.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(getAdHocWsConfigResponse.Error + Environment.NewLine + getAdHocWsConfigResponse.ErrorMsg);
                    return;
                }
                #endregion

                string workstepConfigRecipient1 = getAdHocWsConfigResponse.okInfo.InnerXml;
                XmlDocument wsConfigXmlRecipient1 = new XmlDocument();
                wsConfigXmlRecipient1.LoadXml(workstepConfigRecipient1);

                string envelopeConfigStr = File.ReadAllText(@"..\..\..\Ressources\EnvelopeConfigTemplate_wAuth.xml");
                envelopeConfigStr = envelopeConfigStr.Replace("<!--##WorkstepConfig##-->", workstepConfigRecipient1);

                //adding automatic signature and document timestamp

                // 1 add step in envelope config, recipientType automatic
                // 2 add workstep config with finalize actions for automatic signature
                // 2.1: signature details are defined in signatureTemplate as sig. 
                //      can be either a visible signature (with position) or invisible signature
                //      watch out to have unique id for each signature in document (over the steps and the original document as well)
                // 2.2: define referenced automatic signature profile in sig\AllowedSignatureTypes\sigType\SealingProfileId
                // 3 add timestamp in finalize action of workstep config

                // 1)
                XmlDocument envelopeConfigXml = new XmlDocument();
                envelopeConfigXml.LoadXml(envelopeConfigStr);
                XmlElement stepsElem = (XmlElement)envelopeConfigXml.DocumentElement.SelectSingleNode("steps");

                XmlDocument automaticSigningStepXml = new XmlDocument();
                automaticSigningStepXml.Load(@"..\..\..\Ressources\AutomaticSignature_Step_Template.xml");


                // 2)
                automaticSigningStepXml.DocumentElement.SelectSingleNode("orderIndex").InnerText = "2";

                // 2.2)
                XmlNodeList automaticSignatureProfilesNodesList = automaticSigningStepXml.DocumentElement.SelectNodes("workstepConfiguration/signatureTemplate/sig/AllowedSignatureTypes/sigType/SealingProfileId");
                // in esaw: settings/organization/automatic remote signature profile/(edit) -> identifier
                // my id is : db64a9a6-8cc1-458c-9cbc-b75feb585cf1
                if (automaticSignatureProfilesNodesList != null)
                {
                    for (int i = 0; i < automaticSignatureProfilesNodesList.Count; i++) {
                        automaticSignatureProfilesNodesList[i].InnerText = "db64a9a6-8cc1-458c-9cbc-b75feb585cf1";
                    }
                }

                // 3) --> add FinalizeActionsNode with Timestamp Task
                // this can be added either to the step of the signer, or to the st4p of the automatic signer, or both
                // in this case, I add it to the step of the automatic signer
                // <Policy>
                // <FinalizeActions>
                // <Timestamp DocRefNumber="*" />
                // note: for on premise: 
                //          make sure your default spc uses a configured timestamp, or set a spc that uses a timestamp  <Timestamp DocRefNumber="*" spcId="spcWithTimeStamp"/>

                XmlElement workstepPolicyNodeRecip2 = (XmlElement)automaticSigningStepXml.DocumentElement.SelectSingleNode("workstepConfiguration/Policy");
                XmlElement finalizeActionNodeRecip2 = (XmlElement)workstepPolicyNodeRecip2.SelectSingleNode("FinalizeActions");
                workstepPolicyNodeRecip2.AppendChild(finalizeActionNodeRecip2);

                XmlElement timeStampNodeRecip2 = automaticSigningStepXml.CreateElement("Timestamp");
                XmlAttribute timeStampNodeRecip2DocRefAttrib = automaticSigningStepXml.CreateAttribute("DocRefNumbers");
                timeStampNodeRecip2DocRefAttrib.Value = "*";
                timeStampNodeRecip2.SetAttributeNode(timeStampNodeRecip2DocRefAttrib);
                finalizeActionNodeRecip2.AppendChild(timeStampNodeRecip2);

                // import automatic step into envelopeCongigXml
                stepsElem.AppendChild(envelopeConfigXml.ImportNode(automaticSigningStepXml.DocumentElement, true));    

                string sendEnvelopeResponseXml = ws.SendEnvelope_v1(auth, docIds, envelopeConfigXml.OuterXml);
                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(sendEnvelopeResponseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg, "Error");
                    return;
                }
                Console.WriteLine("EnvelopeÍD: " + response.okInfo.InnerText);

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }

        }

        // this sample shows, how to declare for a signer, that signature fields should be signed with a remote signature or Sms-Otp Signature with predefined values for the signer
        // to be able to use the Sms-Otp signature, you need this feature enabled. Contact our sales team for this.
        //
        // 1) in the envelope configuration, add for the recipients the predefined values: envelope/steps/step/recipients/recipient/
        public static void SendEnvelope_Simple_RemoteSignatureAndSmsOtpSignature_withPredefinedSettingsForSigner(string[] documentsPath)
        {
            try
            {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();

                string[] docIds = UploadPdfsAndGetDocIds(ws, auth, documentsPath);

                #region Analyze Pdf from Server and adjust per recipient
                // this can be ommitted, if you have already a workstep configuration that fits to your document(s) and recipients
                // this code shows, how to programmatically adjust the configuration xml, however, this can be done manually as well and loaded as config

                string adHocConfig = File.ReadAllText(@"..\..\..\Ressources\AdHocConfig.xml");
                string getAdHocWSConfigResponseXml = ws.GetAdHocWorkstepConfiguration_v1(auth, docIds, adHocConfig);

                SignAnywhereResponse getAdHocWsConfigResponse = SignAnywhereResponse.parseResponse(getAdHocWSConfigResponseXml);
                if (getAdHocWsConfigResponse.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(getAdHocWsConfigResponse.Error + Environment.NewLine + getAdHocWsConfigResponse.ErrorMsg);
                    return;
                }
                #endregion

                string workstepConfigRecipient1 = getAdHocWsConfigResponse.okInfo.InnerXml;

                string envelopeConfig = File.ReadAllText(@"..\..\..\Ressources\EnvelopeConfigTemplate_wAuth.xml");
              
                XmlDocument envelopeConfigXml = new XmlDocument();
                envelopeConfigXml.LoadXml(envelopeConfig);


                //1) add predefined values to recipient
                XmlNode firstRecipientNode = envelopeConfigXml.DocumentElement.SelectSingleNode("steps/step/recipients/recipient");
                
                //<smsOtpAdditionalInformation><phoneMobile>+4366412345678</phoneMobile></smsOtpAdditionalInformation>
                XmlElement smsOtpAdditionalInfo = envelopeConfigXml.CreateElement("smsOtpAdditionalInfo");
                smsOtpAdditionalInfo.InnerXml = "<phoneMobile>+436646101700</phoneMobile>";
                firstRecipientNode.AppendChild(smsOtpAdditionalInfo);

                //<remoteCertificateAdditionalInformation> <userId>myUserId</userId><!-- optional --><deviceId>myDeviceId</deviceId></remoteCertificateAdditionalInformation>
                XmlElement remoteCertificateAdditionalInfo = envelopeConfigXml.CreateElement("remoteCertificateAdditionalInformation");
                remoteCertificateAdditionalInfo.InnerXml = "<userId>9aa7ea23</userId><deviceId></deviceId>";
                firstRecipientNode.AppendChild(remoteCertificateAdditionalInfo);

                envelopeConfig = envelopeConfigXml.OuterXml;
                //

                //2) optional: modify workstep config to have only remote Certificate and sms-otp signature as option (for sure this could be done with the workstep adhoc configuration as well

                XmlDocument wsConfigXmlRecipient1 = new XmlDocument();
                wsConfigXmlRecipient1.LoadXml(workstepConfigRecipient1);
                XmlNodeList sigNodes = wsConfigXmlRecipient1.DocumentElement.SelectNodes("signatureTemplate/sig");
                for (int i = 0; i < sigNodes.Count; i++) {
                    XmlElement sigNode = (XmlElement)sigNodes[i];
                    XmlElement allowedSigTypesNode = (XmlElement)sigNode.SelectSingleNode("AllowedSignatureTypes");
                    XmlNodeList allowedSigTypes =   allowedSigTypesNode.SelectNodes("sigType");
                    for (int s = 0; s < allowedSigTypes.Count; s++) {
                        XmlElement allowedSigType = (XmlElement)allowedSigTypes[s];
                        if (allowedSigType.GetAttribute("type") == "TransactionCode" &&
                            (allowedSigType.SelectSingleNode("trModType").InnerText == "RemoteSignature" || allowedSigType.SelectSingleNode("trModType").InnerText == "TransactionCodeSenderPlugin")) { }
                        else
                        {
                            allowedSigTypesNode.RemoveChild(allowedSigType);
                        }
                    }            
                }

                workstepConfigRecipient1 = wsConfigXmlRecipient1.OuterXml;

                envelopeConfig = envelopeConfig.Replace("<!--##WorkstepConfig##-->", workstepConfigRecipient1);

                string sendEnvelopeResponseXml = ws.SendEnvelope_v1(auth, docIds, envelopeConfig);
                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(sendEnvelopeResponseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg, "Error");
                    return;
                }
                Console.WriteLine("EnvelopeÍD: " + response.okInfo.InnerText);

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }

        }


        // this sample shows, how to declare for a recipient, that the signature fields should be signed with a disposable remote certificate
        // to be able to use this feature, you need an LRA Agreement with Namirial. Contact our sales team for this.
        //
        // 1) add additional information for disposable remote certifcate for the recipient to envelope config (<disposableCertificateAdditionalInformation>)
        // 2) in the workstep config for the recipient, adopt the settings for the the signature you want to sign with disposable remote certificate:
        //       a) add sigtype for disposable remote signature to signature
        //       b) Recommended: add qualified timestamp to signature
        // 3) Recommended: in the workstep config for the recipient, add document timestamp

        // 4) optional: group multiple signatures into a single user interaction (batch signing of all signature fields)
        public static void SendEnvelope_Simple_DisposableRemoteCertificate(string[] documentsPath)
        {
            try
            {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();

                string[] docIds = UploadPdfsAndGetDocIds(ws, auth, documentsPath);

                #region Analyze Pdf from Server and adjust per recipient
                // this can be ommitted, if you have already a workstep configuration that fits to your document(s) and recipients
                // this code shows, how to programmatically adjust the configuration xml, however, this can be done manually as well and loaded as config

                string adHocConfig = File.ReadAllText(@"..\..\..\Ressources\AdHocConfig.xml");
                string getAdHocWSConfigResponseXml = ws.GetAdHocWorkstepConfiguration_v1(auth, docIds, adHocConfig);

                SignAnywhereResponse getAdHocWsConfigResponse = SignAnywhereResponse.parseResponse(getAdHocWSConfigResponseXml);
                if (getAdHocWsConfigResponse.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(getAdHocWsConfigResponse.Error + Environment.NewLine + getAdHocWsConfigResponse.ErrorMsg);
                    return;
                }
                #endregion

                string workstepConfigRecipient1 = getAdHocWsConfigResponse.okInfo.InnerXml;                
                XmlDocument wsConfigXmlRecipient1 = new XmlDocument();
                wsConfigXmlRecipient1.LoadXml(workstepConfigRecipient1);                                          
                
                //2
                // SAAS Version (timestamp (2.a) recommanded, but not mandatory)
                // <signatureTemplate>...
                //   <sig id="..." useTimestamp="true"> ...
                //      <AllowedSignatureTypes>   <sigType> ... </sigType> </AllowedSignatureTypes>
                //
                // on premise: 
                // <signatureTemplate>...
                //   <sig id="..."> ... <param name="spcId">spcId1</param></sig>
                //
                //   Recommanded (but not mandatory): use SignaturePluginConfiguration (SPC) that has a timestamp configured (can be be the default spc as well, in that case no param spcID needs to be added)
                XmlNode signatureNodeRecip1 = wsConfigXmlRecipient1.DocumentElement.SelectSingleNode("signatureTemplate");

                string sigTypeTxtDisposable = File.ReadAllText(@"..\..\..\Ressources\SigFieldDefinition_DisposableRemoteSignature_config.xml");
                //string sigTypeTxtDisposable = File.ReadAllText(@"..\..\..\Ressources\SigFieldDefinition_LongLivedDisposableRemoteSignature_config.xml");
                XmlDocument sigTypeTxtDisposableDoc = new XmlDocument();
                sigTypeTxtDisposableDoc.LoadXml(sigTypeTxtDisposable);

                XmlNodeList signatureNodeList = signatureNodeRecip1.SelectNodes("sig");
                if (signatureNodeList != null && signatureNodeList.Count > 0) {
                    foreach (XmlElement sigNode in signatureNodeList)
                    {
                        /* SAAS: 2b: recommended
                         */ 
                         
                        //XmlAttribute useTimeStampAttrib = wsConfigXmlRecipient1.CreateAttribute("useTimestamp");
                        //useTimeStampAttrib.Value = "true";
                        //sigNode.SetAttributeNode(useTimeStampAttrib);

                        XmlNode allowedSigTypeNode = wsConfigXmlRecipient1.ImportNode(sigTypeTxtDisposableDoc.DocumentElement, true);
                        XmlNode allowedSigTypes = sigNode.SelectSingleNode("AllowedSignatureTypes");
                        if (allowedSigTypes == null)
                        {
                            sigNode.AppendChild(allowedSigTypeNode);
                        }
                        else {
                            allowedSigTypes.RemoveAll(); //optional, otherwise the user can select between multiple signature fields                            
                            allowedSigTypes.AppendChild(allowedSigTypeNode.SelectSingleNode("sigType"));
                        }

                    }
                }

                /* 3 --> Recommended: add FinalizeActionsNode with Timestamp Task (can be ommitted as well)
                 * <Policy version="1.1.0.0">
                 * <FinalizeActions>
                 * <Timestamp DocRefNumber="*" />
                 *
                 * note: for on premise: 
                 *          make sure your default spc uses a configured timestamp, or set a spc that uses a timestamp  <Timestamp DocRefNumber="*" spcId="spcWithTimeStamp"/>
                 */

                XmlNode workstepPolicyNodeRecip1 = wsConfigXmlRecipient1.DocumentElement.SelectSingleNode("Policy");

                //XmlElement finalizeActionNodeRecip1 = wsConfigXmlRecipient1.CreateElement("FinalizeActions");
                //workstepPolicyNodeRecip1.AppendChild(finalizeActionNodeRecip1);

                //XmlElement timeStampNodeRecip1 = wsConfigXmlRecipient1.CreateElement("Timestamp");
                //XmlAttribute timeStampNodeRecip1DocRefAttrib = wsConfigXmlRecipient1.CreateAttribute("DocRefNumbers");
                //timeStampNodeRecip1DocRefAttrib.Value = "*";
                //timeStampNodeRecip1.SetAttributeNode(timeStampNodeRecip1DocRefAttrib);

                ////On premise:
                //XmlAttribute timeStampNodeRecip1spcIdAttrib = wsConfigXmlRecipient1.CreateAttribute("spcId");
                //timeStampNodeRecip1spcIdAttrib.Value = "spcWithTimeStamp";
                //timeStampNodeRecip1.SetAttributeNode(timeStampNodeRecip1spcIdAttrib);

                //finalizeActionNodeRecip1.AppendChild(timeStampNodeRecip1);      
          
                /* end of section 3*/

                /* 4 --> fully optional group multiple signatures into a single user interaction (batch signing of all signature fields) 
                 * <Task ... type="SignField" ... batchId="1" />
                 * (this applies to other signature types as well)
                 * (watch out regarding the order of required tasks, that there a no required tasks in between such a batch signing definition)
                 * (you can define multiple batch signing jobs within a workstep, just use another batchId for the additional batch job)
                 */
                XmlNodeList workstepSigningTasksListRecip1 = workstepPolicyNodeRecip1.SelectNodes("WorkstepTasks/Task[@type=\"SignField\"]");
                for (int t = 0; t < workstepSigningTasksListRecip1.Count; t++)
                {
                    XmlElement task = (XmlElement)workstepSigningTasksListRecip1[t];
                    XmlAttribute taskBatchAttrib = wsConfigXmlRecipient1.CreateAttribute("batchId");
                    taskBatchAttrib.Value = "1";
                    task.SetAttributeNode(taskBatchAttrib);
                }
                /* end of section 4*/

                workstepConfigRecipient1 = wsConfigXmlRecipient1.OuterXml;

                string envelopeConfig = File.ReadAllText(@"..\..\..\Ressources\EnvelopeConfigTemplate_DisposableCertificate_FabianWenny_Fake.xml");
                //string envelopeConfig = File.ReadAllText(@"..\..\..\Ressources\EnvelopeConfigTemplate_DisposableCertificate.xml");
                envelopeConfig = envelopeConfig.Replace("<!--##WorkstepConfig##-->", workstepConfigRecipient1);                

                string sendEnvelopeResponseXml = ws.SendEnvelope_v1(auth, docIds, envelopeConfig);
                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(sendEnvelopeResponseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg, "Error");
                    return;
                }
                Console.WriteLine("EnvelopeÍD: " + response.okInfo.InnerText);

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }

        }


        // this sample shows, how you can start an envelope within a POS scenario. The first step will be in a POS scenario, while other steps are as usual with notification to the user via email.
        // in a POS scenario it is not necessary to notify the customer, as he is directly in branch - therefore 
        //    a) disable notification for this step
        //    b) directly open the SAW Viewer or (on premise only) choose one of the other SSP front ends (WSI, Kiosk SDK for SSP, mobile Apps (iOS, Android, Windows Phone/Tablet) for SSP)

        // for simplicity (to keep the code shorter) of the sample, only an envelope with 1 recipient is used, but additional recipients can be added, as demonstrated in sample "SendEnvelope_with2Recipients()"
        // remark: POS scenario step can't be a parallel steps (therefore only one step with order index 1 is allowed)
        public static void SendEnvelope_inPosScenario() {

            try
            {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();

                string[] docIds = UploadPdfsAndGetDocIds(ws, auth, new string[] { @"..\..\..\Ressources\Demo_Contract_Form.pdf" });

                #region Analyze Pdf from Server // this can be ommitted, if you have already a workstep configuration that fits to your document(s)

                string adHocConfig = File.ReadAllText(@"..\..\..\Ressources\AdHocConfig.xml");
                string getAdHocWSConfigResponseXml = ws.GetAdHocWorkstepConfiguration_v1(auth, docIds, adHocConfig);

                SignAnywhereResponse getAdHocWsConfigResponse = SignAnywhereResponse.parseResponse(getAdHocWSConfigResponseXml);
                if (getAdHocWsConfigResponse.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(getAdHocWsConfigResponse.Error + Environment.NewLine + getAdHocWsConfigResponse.ErrorMsg);
                    return;
                }
                #endregion

                //
                string envelopeConfigStr = File.ReadAllText(@"..\..\..\Ressources\EnvelopeConfigTemplate_wAuth.xml");
                envelopeConfigStr = envelopeConfigStr.Replace("<!--##WorkstepConfig##-->", getAdHocWsConfigResponse.okInfo.InnerXml);
                //

                // until here, everyting as in the simple sendEnvelope scenario
                // now adopt the envelope config for this scenario

                XmlDocument envelopeConfigXml = new XmlDocument();
                envelopeConfigXml.LoadXml(envelopeConfigStr);

                //a) disable email sending for the first recipient: add <disableEmail>True</disableEmail>
                XmlElement firstStepElem = (XmlElement)envelopeConfigXml.DocumentElement.SelectSingleNode("steps/step");
                XmlElement firstStepElemRecipientElem = (XmlElement)firstStepElem.SelectSingleNode("recipients/recipient");
                XmlElement disableEmailElem = envelopeConfigXml.CreateElement("disableEmail");
                disableEmailElem.InnerText = XmlConvert.ToString(true);
                firstStepElemRecipientElem.AppendChild(disableEmailElem);

                // optional: delete authentications
                firstStepElem.SelectSingleNode("recipients/recipient/authentications").RemoveAll();
                // optional for SAW Viewer, but IMPORTANT, if other SSP front end then SAW Viewer should be used: 
                //       * adopt workstep config to contain only the supported tasks / features of the selected front end
                //       * watch out for sig types, to use only those, the other front end supports 
                //         (AllowedSignatureTypes notification is most likely not supported by other front ends, check documentation for specific front end) 

                envelopeConfigStr = envelopeConfigXml.OuterXml;

                string sendEnvelopeResponseXml = ws.SendEnvelope_v1(auth, docIds, envelopeConfigStr);
                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(sendEnvelopeResponseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg, "Error");
                    return;
                }
                string envelopeId = response.okInfo.InnerText;
                // string envelopeId = "34262395-22bb-4313-8472-8c663815963c";
                Console.WriteLine("EnvelopeÍD: " + envelopeId);

                // b) open first step of envelope directly in your POS Application
                //     ba) get redirection url
                //     bb) only in case if other front end then SAW viewer should be used (on premise)
                //       bba) get workstep id
                //       bbb) compose url for other front end
                //     bc) open url (either for SAW Viewer or other front end)


                // ba) get redirection url
                //     
                //     //OBSOLETE, if no notification to recipient is sent - in that case, if you have an envelopeID, the redirection url can be directly retrieved from GetEnvelopeStatus
                //     either poll for the envelope status, until status is inProgress
                //     or react on statusUpdateCallback: sendSignNotification
                //

                SignAnywhereEnvelopeStatus envStatus = GetEnvelopeStatus(ws, auth, envelopeId);

                // OBSOlETE
                //while (envStatus.EnvelopeStatus.ToLower() != "inprogress")
                //{
                //    Console.WriteLine("Envelope status is not inProgress, Status is: " + envStatus.EnvelopeStatus);
                //    Thread.Sleep(1000); 
                //    envStatus = GetEnvelopeStatus(ws, auth, envelopeId);
                //}                
                //Console.WriteLine("Envelope status is inProgress");
                string redirectionUrlFirstStep = envStatus.SubEnvelopes[0].Recipients[0].WorkstepRedirectionUrl;

                #region bb) optional: to use another front end
                // bb) optional: to use another front end
                // bba) get workstepId
                string requestUrlForWorkstepId = redirectionUrlFirstStep + "&responseType=returnWorkstepId";
                HttpWebRequest webRequest = (HttpWebRequest)HttpWebRequest.Create(requestUrlForWorkstepId);
                webRequest.AllowAutoRedirect = false;
                HttpWebResponse workstepIdResponse = (HttpWebResponse)webRequest.GetResponse();
                if (workstepIdResponse.StatusCode != HttpStatusCode.OK)
                {
                    Console.WriteLine("Invalid status code returned from Server: " + workstepIdResponse.StatusCode);
                    return;
                }
                Stream responseStream = workstepIdResponse.GetResponseStream();
                StreamReader reader = new StreamReader(responseStream);
                string workstepId = reader.ReadToEnd();

                // bbc) compose url for other front end
                //      see integration guide for SSP and related front ends
                redirectionUrlFirstStep = "http://beta4.testlab.xyzmo.com:50100/Sign.aspx?WorkstepID=" + workstepId;
                #endregion

                //bc) open url
                System.Diagnostics.Process.Start(redirectionUrlFirstStep);


            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }

        }

        #endregion

        #region GetEnvelopeStatus

        private static SignAnywhereEnvelopeStatus GetEnvelopeStatus(eSAW.Api ws, string auth,string envelopeId)
        {
                string envelopeStatusResponseXml = ws.GetEnvelopeById_v1(auth, envelopeId);
                SignAnywhereResponse envStatusResponse = SignAnywhereResponse.parseResponse(envelopeStatusResponseXml);

                if (envStatusResponse.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    
                    string errorMsg = "Error on retrieving envelope status:" + Environment.NewLine +
                        envStatusResponse.Error + Environment.NewLine + envStatusResponse.ErrorMsg;
                    throw new ApplicationException(errorMsg);
                }

                SignAnywhereEnvelopeStatus envStatus = SignAnywhereEnvelopeStatus.FromXmlElem((XmlElement)envStatusResponse.okInfo.SelectSingleNode("envelopeStatus"));
                return envStatus;
        }

        #endregion

        #region Creating Drafts


        public static void CreateDraft(string[] documentPaths)
        {
            try
            {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();

                string createDraftConfigStr = File.ReadAllText(@"..\..\..\Ressources\EnvelopeConfig_forDraft_wAuth.xml");
                //string createDraftConfigStr = File.ReadAllText(@"..\..\..\Ressources\EnvelopeConfig_forDraft_Disposable.xml");

                //define the page of the design step the user is redirected when the draft is opened to be continued: ToDesigner, ToRecipients, ToSend, None
                //define the redirection url for the user, when he clicks on sending the envelope
                //define the callback url to notify your back end, that the user sent this envelope. eg: http://172.16.17.78:57550/default.aspx?EnvelopeId=##EnvelopeId##&action=##Action##&CallbackFromCreateDraft=1234 
                //(make sure to escape the & with &amp; , when you place this url in the draftOptions xml)
                //
                //the design step can be even embedded into an iFrame (only in AgentRedirect Mode, see below). Just add a white list for allow top frames <iFrameWhiteList>http://172.16.17.181/iframe/;http://foo.org</iFrameWhiteList>
                string redirectionConfig = "<draftOptions><afterSendRedirectUrl>http://www.xyzmo.com</afterSendRedirectUrl><afterSendCallbackUrl></afterSendCallbackUrl><redirectPolicy>ToRecipients</redirectPolicy><allowAgentRedirect>true</allowAgentRedirect></draftOptions>";

                string[] docIds = UploadPdfsAndGetDocIds(ws, auth, documentPaths);

                string createDraftResponseXml = ws.CreateDraft_v1(auth, docIds, createDraftConfigStr, redirectionConfig);
                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(createDraftResponseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Ok)
                {
                    Console.WriteLine("Draft created with id: " + response.okInfo.InnerText);

                    // now you can redirect the user to continue the draft within eSignAnyWhere:
                    
                    // use this link for on premise in the non  AD/LDAP version or on www.significant.com: /AgentRedirect/index?draftid=      (draft can be accessed without login)
                    // in the draftOptions, add <allowAgentRedirect>true</allowAgentRedirect>
                    System.Diagnostics.Process.Start(eSawUrl + @"/AgentRedirect/index?draftid=" + response.okInfo.InnerText);
                    
                    // use this link on www.significant.com or for on premise with AD / LDAP version: /redirect/toDefinition?Id=              (requires the user to login)
                    //System.Diagnostics.Process.Start(eSawUrl + @"/redirect/toDefinition?Id=" + response.okInfo.InnerText);

                }
                else
                {
                    Console.WriteLine (response.Error + Environment.NewLine + response.ErrorMsg);
                }

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }



        #endregion


        #region Working with Templates

        public static void SendEnvelopeFromTemplate(string templateEnvelopeID) {

            try
            {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();

                string templateInformationResponseXml = ws.GetEnvelopeById_v1(auth, templateEnvelopeID);

                string xmlOverrideOptions = File.ReadAllText(@"..\..\..\Ressources\TemplateOverrideOptions.xml");

                string sendEnvelopeResponseXml = ws.SendEnvelopeFromTemplate_v1(auth, templateEnvelopeID, xmlOverrideOptions);

                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(sendEnvelopeResponseXml);               
                if (response.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg, "Error");
                    return;
                }
                Console.WriteLine("EnvelopeÍD: " + response.okInfo.InnerText);

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }

        }


        public static void CreateDraftFromTemplate(string TemplateEnvelopeID) {
            try
            {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();

                string redirectionConfig = "<draftOptions><afterSendRedirectUrl>http://www.xyzmo.com</afterSendRedirectUrl><afterSendCallbackUrl>http://172.16.17.138:57550/default.aspx?EnvelopeId=##DraftId##&amp;CallbackFromCreateDraft=1234</afterSendCallbackUrl><redirectPolicy>ToDesigner</redirectPolicy></draftOptions>";

                string xmlOverrideOptions = File.ReadAllText(@"..\..\..\Ressources\TemplateOverrideOptions.xml");

                string createDraftResponseXml = ws.CreateDraftFromTemplate_v1(auth, TemplateEnvelopeID, xmlOverrideOptions, redirectionConfig);
                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(createDraftResponseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg, "Error");
                    return;
                }
                Console.WriteLine("EnvelopeÍD: " + response.okInfo.InnerText);

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }

        }

        
        // deprecated method of working with templates. use SendEnvelopeFromTemplate instead
        public static void UsingATemplateForSendEnvelope(string templateEnvelopeID)
        {

            try
            {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();

                string templateInformationResponseXml = ws.GetEnvelopeById_v1(auth, templateEnvelopeID);
                SignAnywhereResponse templateInformationResponse = SignAnywhereResponse.parseResponse(templateInformationResponseXml);
                if (templateInformationResponse.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(templateInformationResponse.Error + Environment.NewLine + templateInformationResponse.ErrorMsg, "Error");
                    return;
                }
                Console.WriteLine("\t" + DateTime.Now.ToLongTimeString() + " GetTemplateInfo success.");

                SignAnywhereEnvelopeStatus envStatus = SignAnywhereEnvelopeStatus.FromXmlElem((XmlElement)templateInformationResponse.okInfo.SelectSingleNode("envelopeStatus"));
                //string xmlOverrideOptions = File.ReadAllText(@"..\..\..\Ressources\TemplateOverrideOptions.xml");

                string copyEnvelopeResponseXml = ws.CopyDocumentFromTemplate_v1(auth, templateEnvelopeID);
                SignAnywhereResponse copyEnvelopeResponse = SignAnywhereResponse.parseResponse(copyEnvelopeResponseXml);
                if (copyEnvelopeResponse.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(copyEnvelopeResponse.Error + Environment.NewLine + copyEnvelopeResponse.ErrorMsg, "Error");
                    return;
                }
                Console.WriteLine("\t" + DateTime.Now.ToLongTimeString() + " CopyTemplate success.");
                /*
                < CopyDocumentFromTemplateResult >
                    < SspFileIds >
                        < SspFileId > 39a24116 - f9f3 - 4e8a - ac80 - 11477a9d5319 </ SspFileId >           
                    </ SspFileIds >           
                    < envelope > ... envelopeDescriptionXml
                    </envelope>

                */
                XmlNodeList sspFileIdsList = copyEnvelopeResponse.okInfo.SelectNodes("CopyDocumentFromTemplateResult/SspFileIds/SspFileId");
                string[] sspFileIds = new string[sspFileIdsList.Count];
                for (int i = 0; i < sspFileIdsList.Count; i++)
                {
                    sspFileIds[i] = sspFileIdsList[i].InnerText;
                }
                //adjust sendEnvelope Description
                XmlElement sendEnvelopeDescXmlElem = (XmlElement)copyEnvelopeResponse.okInfo.SelectSingleNode("CopyDocumentFromTemplateResult/envelope");

                // override envelope settings
                sendEnvelopeDescXmlElem.SelectSingleNode("name").InnerText = "My new name for the envelope";

                // replace recipients
                // all recipients can be replaced, placeholder recipient (email starts with "placeholder:" must be replaced)
                XmlNodeList recipientsNodeList = sendEnvelopeDescXmlElem.SelectNodes("steps/step/recipients/recipient");
                for (int i = 0; i < recipientsNodeList.Count; i++)
                {
                    recipientsNodeList[i].SelectSingleNode("eMail").InnerText = "a.barletta@namirial.com";
                    recipientsNodeList[i].SelectSingleNode("firstName").InnerText = "new first name";
                    recipientsNodeList[i].SelectSingleNode("lastName").InnerText = "new last name";
                    recipientsNodeList[i].SelectSingleNode("languageCode").InnerText = "en";
                }

                // optional: prefill form field values
                XmlDocument preFillFormFieldsXmlDoc = new XmlDocument();
                //preFillFormFieldsXmlDoc.Load(@"..\..\..\Ressources\Demo_Contract_Form_OverrideFormFieldValues.xml");
                preFillFormFieldsXmlDoc.Load(@"..\..\..\Ressources\Demo_Contract_Form_OverrideFormFieldValues - Copy.xml");

                sendEnvelopeDescXmlElem.AppendChild(sendEnvelopeDescXmlElem.OwnerDocument.ImportNode(preFillFormFieldsXmlDoc.DocumentElement, true));

                string envelopeDescriptionXml = sendEnvelopeDescXmlElem.OuterXml;

                string sendEnvelopeResponseXml = ws.SendEnvelope_v1(auth, sspFileIds, envelopeDescriptionXml);

                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(sendEnvelopeResponseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg, "Error");
                    return;
                }
                Console.WriteLine("\t" + DateTime.Now.ToLongTimeString() + " SendEnvelope success.");
                Console.WriteLine("EnvelopeÍD: " + response.okInfo.InnerText);

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }

        }

        #endregion

        #region Managing Envelops and Drafts

        public static void SendReminder(string envelopeId)
        {
            try
            {
                if (String.IsNullOrWhiteSpace(envelopeId))
                {
                    Console.WriteLine("Workflow Id must be set.");
                    return;
                }

                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();
                string responseXml = ws.SendReminders_v1(auth, envelopeId);
                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(responseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }


        public static void CancelEnvelope(string envelopeId)
        {
            try
            {
                if (String.IsNullOrWhiteSpace(envelopeId))
                {
                    Console.WriteLine("Workflow Id must be set.");
                    return;
                }

                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();
                string responseXml = ws.CancelEnvelope_v1(auth, envelopeId);
                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(responseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }

        public static void RestartEnvelope(string envelopeId)
        {
            try
            {
                if (String.IsNullOrWhiteSpace(envelopeId))
                {
                    Console.WriteLine("Workflow Id must be set.");
                    return;
                }

                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();
                string responseXml = ws.RestartEnvelope_v1(auth, envelopeId,5);
                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(responseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }

        public static void DeleteEnvelope(string envelopeId)
        {
            try
            {
                if (String.IsNullOrWhiteSpace(envelopeId))
                {
                    Console.WriteLine("Workflow Id must be set.");
                    return;
                }

                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();
                string responseXml = ws.DeleteEnvelope_v1(auth, envelopeId);

                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(responseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }

        #endregion

        #region Download Completed Documents

        public static void DownloadCompletedEnvelope(string envelopeID)
        {
            try
            {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();

                    string envelopeStatusResponseXml = ws.GetEnvelopeById_v1(auth, envelopeID);
                SignAnywhereResponse envStatusResponse = SignAnywhereResponse.parseResponse(envelopeStatusResponseXml);

                if (envStatusResponse.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine("Error on retrieving envelope status:" + Environment.NewLine +
                        envStatusResponse.Error + Environment.NewLine + envStatusResponse.ErrorMsg, "Error");
                    return;
                }

                SignAnywhereEnvelopeStatus envStatus = SignAnywhereEnvelopeStatus.FromXmlElem((XmlElement)envStatusResponse.okInfo.SelectSingleNode("envelopeStatus"));

                if (envStatus.EnvelopeStatus.ToLower() != "completed" && envStatus.EnvelopeStatus.ToLower() != "bulkpartlycompleted" && envStatus.EnvelopeStatus.ToLower() != "bulkcompleted")
                {
                    Console.WriteLine("Envelope status is not completed, Status is: " + envStatus.EnvelopeStatus);
                    return;
                }

                if (!String.IsNullOrWhiteSpace(envStatus.MetaData))
                {
                    Console.WriteLine("Metadata: " + envStatus.MetaData);
                }
                else {
                    Console.WriteLine("No metadata for this envelope set.");
                }

                DirectoryInfo di = System.IO.Directory.CreateDirectory(@"..\..\..\output\" + envelopeID);
                Console.WriteLine("Writing files to " + di.FullName);

                for (int i = 0; i < envStatus.SubEnvelopes.Count; i++)
                {
                    if (envStatus.SubEnvelopes[i].Status.ToLower() == "completed") // for 3.1 and bulk envelopes: check status for sub envelope too ...
                    {
                        DirectoryInfo diForSubEnv = di;
                        if (envStatus.SubEnvelopes.Count > 1)
                        {
                            diForSubEnv = new DirectoryInfo(di.FullName + "\\" + envStatus.SubEnvelopes[i].RecipientEmail.Replace("@", "(at)"));
                            diForSubEnv.Create();
                        }

                        for (int f = 0; f < envStatus.SubEnvelopes[i].FinishedDocuments.DocumentList.Count; f++)
                        {
                            string docName = envStatus.SubEnvelopes[i].FinishedDocuments.DocumentList[f].DocumentName;
                            Console.WriteLine(docName);
                            if (envStatus.SubEnvelopes[i].FinishedDocuments.DocumentList[f].FormFieldValues == null || envStatus.SubEnvelopes[i].FinishedDocuments.DocumentList[f].FormFieldValues.FormFieldValues.Count <= 0)
                            {
                                Console.WriteLine("No form fields for this document!");
                            }
                            else
                            {
                                for (int ff = 0; ff < envStatus.SubEnvelopes[i].FinishedDocuments.DocumentList[f].FormFieldValues.FormFieldValues.Count; ff++)
                                {
                                    Console.WriteLine("\t" + envStatus.SubEnvelopes[i].FinishedDocuments.DocumentList[f].FormFieldValues.FormFieldValues[ff].Key + ": " + envStatus.SubEnvelopes[i].FinishedDocuments.DocumentList[f].FormFieldValues.FormFieldValues[ff].Value);
                                }
                            }

                            SignAnywhereEnvelopeStatus.Documents.CompletedDocumentInfo docInfo = envStatus.SubEnvelopes[i].FinishedDocuments.DocumentList[f];
                            

                            string fileDownLoadResponseXml = ws.DownloadCompletedDocument_v1(auth, envelopeID, envStatus.SubEnvelopes[i].FinishedDocuments.DocumentList[f].DocumentId);
                            SignAnywhereResponse fileDownLoadResponse = SignAnywhereResponse.parseResponse(fileDownLoadResponseXml);
                            if (fileDownLoadResponse.BaseResult == SignAnywhereResponse.Result.Ok)
                            {
                                SignAnywhereFile file = SignAnywhereFile.FromXmlElement((XmlElement)fileDownLoadResponse.okInfo.SelectSingleNode("file"));
                                string filePath = diForSubEnv.FullName + @"\" + file.FileName;
                                File.WriteAllBytes(filePath, file.Data);
                                Console.WriteLine("Document downloaded to: " + filePath);
                            }
                            else
                            {
                                Console.WriteLine(fileDownLoadResponse.Error + Environment.NewLine + fileDownLoadResponse.ErrorMsg);
                            }

                            if (docInfo.Attachments != null && docInfo.Attachments.Count > 0)
                            {
                                for (int a = 0; a < docInfo.Attachments.Count; a++)
                                {
                                    string attachmentDownLoadResponseXml = ws.DownloadCompletedDocument_v1(auth, envelopeID, docInfo.Attachments[a].DocumentId);
                                    SignAnywhereResponse attachmentDownLoadResponse = SignAnywhereResponse.parseResponse(attachmentDownLoadResponseXml);
                                    if (attachmentDownLoadResponse.BaseResult == SignAnywhereResponse.Result.Ok)
                                    {
                                        SignAnywhereFile file = SignAnywhereFile.FromXmlElement((XmlElement)attachmentDownLoadResponse.okInfo.SelectSingleNode("file"));
                                        string filePath = diForSubEnv.FullName + @"\" + docInfo.DocumentName+ "__"+ file.FileName;
                                        File.WriteAllBytes(filePath, file.Data);
                                        Console.WriteLine("Attachment downloaded to: " + filePath);
                                    }
                                    else
                                    {
                                        Console.WriteLine(attachmentDownLoadResponse.Error + Environment.NewLine + attachmentDownLoadResponse.ErrorMsg);
                                    }
                                }
                            }
                            else
                            {
                                Console.WriteLine("No attachments in this document!");
                            }

                            // log per document (optional)
                            if (!String.IsNullOrWhiteSpace(envStatus.SubEnvelopes[i].FinishedDocuments.DocumentList[f].LogDocIdPerDoc))
                            {
                                string logFileDownLoadResponseXml = ws.DownloadCompletedDocument_v1(auth, envelopeID, envStatus.SubEnvelopes[i].FinishedDocuments.DocumentList[f].LogDocIdPerDoc);
                                SignAnywhereResponse logFileDownLoadResponse = SignAnywhereResponse.parseResponse(logFileDownLoadResponseXml);
                                if (logFileDownLoadResponse.BaseResult == SignAnywhereResponse.Result.Ok)
                                {
                                    SignAnywhereFile file = SignAnywhereFile.FromXmlElement((XmlElement)logFileDownLoadResponse.okInfo.SelectSingleNode("file"));
                                    string filePath = diForSubEnv.FullName + @"\" + file.FileName;
                                    File.WriteAllBytes(filePath, file.Data);
                                    Console.WriteLine("LogPerDocument downloaded to: " + filePath);
                                }
                                else
                                {
                                    Console.WriteLine(fileDownLoadResponse.Error + Environment.NewLine + fileDownLoadResponse.ErrorMsg);
                                }
                            }

                            // images per document - 
                            for (int docs = 0; docs < envStatus.EnvelopeDocuments.Count; docs++)
                            {
                                for (int pageNumberIndex = 0; pageNumberIndex < envStatus.EnvelopeDocuments[docs].Pages.Count; pageNumberIndex++)
                                {
                                    string envId = String.IsNullOrWhiteSpace(envStatus.SubEnvelopes[i].SubEnvelopeId)? envelopeID: envStatus.SubEnvelopes[i].SubEnvelopeId;
                                    string pageDownLoadResponseXml = ws.DownloadEnvelopePageImage_v1(auth, envId , envStatus.EnvelopeDocuments[docs].WorkstepDocRefNumber.ToString() , (pageNumberIndex + 1).ToString());
                                    SignAnywhereResponse pageDownLoadResponse = SignAnywhereResponse.parseResponse(pageDownLoadResponseXml);
                                    if (pageDownLoadResponse.BaseResult == SignAnywhereResponse.Result.Ok)
                                    {
                                        SignAnywhereFile file = SignAnywhereFile.FromXmlElement((XmlElement)pageDownLoadResponse.okInfo.SelectSingleNode("file"));
                                        string filePath = diForSubEnv.FullName + @"\" + file.FileName;
                                        File.WriteAllBytes(filePath, file.Data);
                                        Console.WriteLine("Image of page downloaded to: " + filePath);
                                    }
                                    else
                                    {
                                        Console.WriteLine(pageDownLoadResponse.Error + Environment.NewLine + pageDownLoadResponse.ErrorMsg);
                                    }
                                }
                            }

                        }
                        // log
                        string envelopeLogDownloadResponseXml = ws.DownloadCompletedDocument_v1(auth, envelopeID, envStatus.SubEnvelopes[i].FinishedDocuments.LogDocId);
                        SignAnywhereResponse envelopeLogDownloadResponse = SignAnywhereResponse.parseResponse(envelopeLogDownloadResponseXml);
                        if (envelopeLogDownloadResponse.BaseResult == SignAnywhereResponse.Result.Ok)
                        {
                            SignAnywhereFile file = SignAnywhereFile.FromXmlElement((XmlElement)envelopeLogDownloadResponse.okInfo.SelectSingleNode("file"));
                            File.WriteAllBytes(diForSubEnv.FullName + @"\" + file.FileName, file.Data);
                        }
                        else
                        {
                            Console.WriteLine(envelopeLogDownloadResponse.Error + Environment.NewLine + envelopeLogDownloadResponse.ErrorMsg);
                        }

                        // logXml
                        string envelopeLogXmlDownloadResponseXml = ws.DownloadCompletedDocument_v1(auth, envelopeID, envStatus.SubEnvelopes[i].FinishedDocuments.LogXmlDocId);
                        SignAnywhereResponse envelopeLogXmlDownloadResponse = SignAnywhereResponse.parseResponse(envelopeLogXmlDownloadResponseXml);
                        if (envelopeLogXmlDownloadResponse.BaseResult == SignAnywhereResponse.Result.Ok)
                        {
                            SignAnywhereFile file = SignAnywhereFile.FromXmlElement((XmlElement)envelopeLogXmlDownloadResponse.okInfo.SelectSingleNode("file"));
                            File.WriteAllBytes(diForSubEnv.FullName + @"\" + file.FileName, file.Data);
                        }
                        else
                        {
                            Console.WriteLine(envelopeLogDownloadResponse.Error + Environment.NewLine + envelopeLogDownloadResponse.ErrorMsg);
                        }
                    }
                }
                Console.WriteLine("All Documents have been downloaded successfully!");
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString(), "Error");
            }
        }

        #endregion

        public static void GetEnvelopeStatus(string envelopeId) { 
        
         try
            {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();

                string envelopeStatusResponseXml = ws.GetEnvelopeById_v1(auth, envelopeId);
                SignAnywhereResponse envStatusResponse = SignAnywhereResponse.parseResponse(envelopeStatusResponseXml);

                if (envStatusResponse.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine("Error on retrieving envelope status:" + Environment.NewLine +
                        envStatusResponse.Error + Environment.NewLine + envStatusResponse.ErrorMsg, "Error");
                    return;
                }

                SignAnywhereEnvelopeStatus envStatus = SignAnywhereEnvelopeStatus.FromXmlElem((XmlElement)envStatusResponse.okInfo.SelectSingleNode("envelopeStatus"));
                Console.WriteLine("Status for envelope id " + envelopeId + ": " + envStatus.EnvelopeStatus); 

            }
         catch (Exception ex)
         {
             Console.WriteLine(ex.ToString(), "Error");
         }
        }


        #region FindingEnvelops

        
        // this methods lets you find all envelops for a user. use this method with care
        // you can even look for all templates that are visible to this user and use one of the templates to send an envelope based on this template (see SendEnvelopeFromTemplate() )
        public static void FindEnvelops() {

            try
            {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();

                string findEnvelopeDescriptionXml = "<findEnvelopesDescriptor><status>Template</status></findEnvelopesDescriptor>";
                // values for Status: Draft|Started|InProgress|Canceled|Completed|Expired|Rejected|Template|ActionRequired|WaitingForOthers|ExpiringSoon|Active
                // example: "<findEnvelopesDescriptor><status>Completed</status></findEnvelopesDescriptor>";
                //
                // in addition, you can look up for envelopes, a certain recipient is assigned
                // <findEnvelopesDescriptor><status></status><hasRecipient>some@mail.com</hasRecipient></findEnvelopesDescriptor>

                /*
                 * <findEnvelopesDescriptor> <!-- all search parameters are optional -->
                         <status>Draft|Started|InProgress|Canceled|Completed|Expired|Rejected|Template|ActionRequired|WaitingForOthers|ExpiringSoon|Active</status>
                         <bulk>dc4cdaa9-c204-470f-986d-94786ff159f7</bulk> <!-- 3.1: new filter parameter for bulk, to get all sub envelopes of this bulk envelope-->
                         <inStatusSinceDays>30</inStatusSinceDays>
                         <sentAfterDate>2017-02-08T12:18:37.2415657Z</sentAfterDate> <!-- all dates can also be "date only" -->
                         <sentBeforeDate>2017-05-10T11:18:37.2415657Z</sentBeforeDate>
                         <searchText>some text</searchText>
                         <hasSender>
                            <eMail>some@mail.com</eMail>
                         </hasSender>
                         <hasSigner>
                            <eMail>some@mail.com</eMail>
                         </hasSigner>
                         <hasRecipient>
                            <eMail>some@mail.com</eMail>  <!-- if just one recipient is searched for, you can leave the "eMail" tag because of backwards compatibility -->
                         </hasRecipient>
                         <waitingForRecipient>some@mail.com</waitingForRecipient>
                     </findEnvelopesDescriptor>
                  * 
                  */
                //string findEnvelopeDescriptionXml = "<findEnvelopesDescriptor><status>Completed</status><sentAfterDate>2017-04-30</sentAfterDate></findEnvelopesDescriptor>";
                
                // resultFormat: empty, "OnlyIdentifiers", "Extended" 

                // result on "Extended":
                /*
                 * <findEnvelopesResult>
                      <envelope>
                        <id>28e487cb-d5fa-4651-8810-2f28dd319ef9</id>
                        <name></name>
                        <status>Draft</status>
                      </envelope>
                      <envelope>
                        <id>9be32745-5c5f-4e27-b1fd-708f86bd1ffb</id>
                        <name>DemoContract_Basic</name>
                        <status>Template</status>                       
                      </envelope>
                      <envelope>
                        <id>9be32745-5c5f-4e27-a5c7-708f86bd1ffb</id>
                        <name>DemoContract_Basic</name>
                        <status>Completed</status>    
                        <bulk>dc4cdaa9-c204-470f-986d-94786ff159f7</bulk> <!-- 3.1: in case of a bulk, the id of the parent envelope is returned -->
                      </envelope>
                 * </findEnvelopesResult>
                 */

                string responseXml = ws.FindEnvelopes_v2(auth, findEnvelopeDescriptionXml, "Extended");

                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(responseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg, "Error");
                    return;
                }
                
                Console.WriteLine("EnvelopeÍD: " + response.okInfo.InnerXml);
                //get detailed status for each envelope with "GetEnvelopeById_v1". See above for details

            }

            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }



        #endregion

        public static void ReplaceOrChangeRecipientInEnvelope()
        {
            try {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();

                string recipientOverrideXml = File.ReadAllText(@"..\..\..\Ressources\RecipientOverride.xml");
                recipientOverrideXml = recipientOverrideXml.Replace("##recipLookupEmail##", "rd@xyzmo.com");
                recipientOverrideXml = recipientOverrideXml.Replace("##orderIndex##", XmlConvert.ToString(1));
                string repRecipResponseXml = ws.ReplaceRecipient_v1(auth, "aaec8843-29da-43d8-8251-c035f18b0c1f", recipientOverrideXml);

                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(repRecipResponseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Ok)
                {
                    Console.WriteLine("Recipient successfully changed or replaced in envelope.");

                }
                else
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }


        public static void ChangeRecipientsTasksInEnvelope()
        {
            try
            {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();

                // send an envelope (same code as sendEnvelopeSimple
                #region sendEnvelope
                string[] docIds = UploadPdfsAndGetDocIds(ws, auth, new string[] { @"..\..\..\Ressources\Demo_Contract_Form.pdf" });

                #region Analyze Pdf from Server
                // this can be ommitted, if you have already a workstep configuration that fits to your document(s)

                string adHocConfig = File.ReadAllText(@"..\..\..\Ressources\AdHocConfig.xml");
                string getAdHocWSConfigResponseXml = ws.GetAdHocWorkstepConfiguration_v1(auth, docIds, adHocConfig);

                SignAnywhereResponse getAdHocWsConfigResponse = SignAnywhereResponse.parseResponse(getAdHocWSConfigResponseXml);
                if (getAdHocWsConfigResponse.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(getAdHocWsConfigResponse.Error + Environment.NewLine + getAdHocWsConfigResponse.ErrorMsg);
                    return;
                }

                // note for on premise versions: make sure to have the correct signing certificate configured in your workstep configuration.
                //                               workstepConfiguration/signaturePluginConfiguration/PdfSignatureCryptographicData_V1/SigningCertificateDescriptor/Identifier
                #endregion //analyze

                //
                string envelopeConfig = File.ReadAllText(@"..\..\..\Ressources\EnvelopeConfigTemplate_wAuth.xml");

                string workstepConfig = getAdHocWsConfigResponse.okInfo.InnerXml;

                envelopeConfig = envelopeConfig.Replace("<!--##WorkstepConfig##-->", workstepConfig);


                string sendEnvelopeResponseXml = ws.SendEnvelope_v1(auth, docIds, envelopeConfig);
                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(sendEnvelopeResponseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Failed)
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg, "Error");
                    return;
                }
                string envelopeId = response.okInfo.InnerText;
                TimeSpan wait = new TimeSpan(0, 2, 0);
                Console.WriteLine("Envelope successfull sent. ID: " + envelopeId);

                // end sendEnvelope
                #endregion //sendEnvelope

                Console.WriteLine("Just for the possibility to open the first step, waiting now for " + wait.TotalSeconds + " seconds .....");
                Thread.Sleep(wait);
                //

                string recipientOverrideXml = File.ReadAllText(@"..\..\..\Ressources\RecipientOverride.xml");
                recipientOverrideXml = recipientOverrideXml.Replace("##recipLookupEmail##", "signer.one@xyzmo.com");
                recipientOverrideXml = recipientOverrideXml.Replace("##orderIndex##", XmlConvert.ToString(1));

                XmlDocument recipOverrideXmlDoc = new XmlDocument();
                recipOverrideXmlDoc.LoadXml(recipientOverrideXml);
                // optional: remove recipient node, as I do not want the change the recipient
                XmlNode newOverrideRecipNode = recipOverrideXmlDoc.DocumentElement.SelectSingleNode("recipient");                
                recipOverrideXmlDoc.DocumentElement.RemoveChild(newOverrideRecipNode);

                //adjust workstep config
                XmlDocument workstepConfigXmlDoc = new XmlDocument();
                workstepConfigXmlDoc.LoadXml(workstepConfig);

                XmlNode signatureNode = workstepConfigXmlDoc.DocumentElement.SelectSingleNode("signatureTemplate");
                XmlNode workstepTasksNode = workstepConfigXmlDoc.DocumentElement.SelectSingleNode("Policy/WorkstepTasks");

                signatureNode.RemoveChild(signatureNode.SelectSingleNode("sig[@id=\"Signature13\"]"));
                workstepTasksNode.RemoveChild(workstepTasksNode.SelectSingleNode("Task[@id=\"Signature13\"]"));

                //add modified workstepconfig to recipient override Xml
                XmlNode workstepConfigInRecipOverrideXml = recipOverrideXmlDoc.ImportNode(workstepConfigXmlDoc.DocumentElement, true);
                recipOverrideXmlDoc.DocumentElement.AppendChild(workstepConfigInRecipOverrideXml);
                //

                string repRecipResponseXml = ws.ReplaceRecipient_v1(auth, envelopeId, recipOverrideXmlDoc.OuterXml);

                SignAnywhereResponse repResponse = SignAnywhereResponse.parseResponse(repRecipResponseXml);
                if (repResponse.BaseResult == SignAnywhereResponse.Result.Ok)
                {
                    Console.WriteLine("WorkstepConfig successfully changed for selected recipient in envelope.");
                }
                else
                {
                    Console.WriteLine(repResponse.Error + Environment.NewLine + repResponse.ErrorMsg);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }



        public static void DeleteRecipientFromEnvelope() {

            try
            {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();


                string recipientLookUpXml = "<recipientLookUp><eMail>address.to@lookup.com</eMail><orderIndex>1</orderIndex></recipientLookUp>";
                string delRecipResponseXml = ws.DeleteRecipient_v1(auth, "1e7c2521-d2a7-4df6-b460-168424303cf8", recipientLookUpXml);

                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(delRecipResponseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Ok)
                {
                    Console.WriteLine("Recipient successfully deleted from envelope.");

                }
                else
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg);
                }

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        
        }

        public static void ValidateAuthorization()
        {
            try
            {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();

                string valAuthResponseXml = ws.ValidateAuthorization_v1(auth);

                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(valAuthResponseXml);
                if (response.BaseResult == SignAnywhereResponse.Result.Ok)
                {
                    Console.WriteLine("Authorization valid");

                }
                else
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg);
                }

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }

        public static void FindAllEnvelopsThatAreWaitingForDedicatedRecipientAndReplaceRecipient()
        {
            // this is useful, in case a signer is not available and someone else should sign instead of this signer (employee is on vacation, vacation replacement should sign instead)

            try
            {
                eSAW.Api ws = GetSawWebService();
                string auth = GetApiAuthorization();

                // currently waiting for this recipient
                string recipEmail = "recipient@email.com";
                 string findEnvDescr = "<findEnvelopesDescriptor><status></status><waitingForRecipient>"+recipEmail+"</waitingForRecipient></findEnvelopesDescriptor>";
                
                string findResponseXml = ws.FindEnvelopes_v1(auth, findEnvDescr);

                SignAnywhereResponse findResponse = SignAnywhereResponse.parseResponse(findResponseXml);
                if (findResponse.BaseResult != SignAnywhereResponse.Result.Ok)                
                {
                    Console.WriteLine(findResponse.Error + Environment.NewLine + findResponse.ErrorMsg);
                    return;
                }
                //<findEnvelopesResult><envelopeId>1c627033-19c5-4b15-8f13-983772ff7210</envelopeId></findEnvelopesResult>
                XmlNodeList envelopeList =  findResponse.okInfo.ChildNodes[0].SelectNodes("envelopeId");
                if (envelopeList == null || envelopeList.Count == 0)
                {
                    Console.WriteLine("No envelopes found for recipient");
                }
                else
                {
                    for (int i = 0; i < envelopeList.Count; i++) {
                        Console.WriteLine(envelopeList[i].InnerText);
                        
                        // make sure to have the order index of the recipient to replace set appropriate (use GetEnvelopeById_v1 to be sure about the orderIndex)
                        string envelopeStatusResponseXml = ws.GetEnvelopeById_v1(auth, envelopeList[i].InnerText);
                        SignAnywhereResponse envStatusResponse = SignAnywhereResponse.parseResponse(envelopeStatusResponseXml);

                        if (envStatusResponse.BaseResult == SignAnywhereResponse.Result.Failed)
                        {
                            Console.WriteLine("Error on retrieving envelope status:" + Environment.NewLine +
                                envStatusResponse.Error + Environment.NewLine + envStatusResponse.ErrorMsg, "Error");
                            return;
                        }
                        SignAnywhereEnvelopeStatus envStatus = SignAnywhereEnvelopeStatus.FromXmlElem((XmlElement)envStatusResponse.okInfo.SelectSingleNode("envelopeStatus"));
                        int orderIndex = -1;
                        int r = 0;
                        while (orderIndex == -1 && r < envStatus.SubEnvelopes[0].Recipients.Count)
                        {
                            SignAnywhereEnvelopeStatus.Recipient envRec = envStatus.SubEnvelopes[0].Recipients[r];
                            if (envRec.Status.Equals("notsigned", StringComparison.CurrentCultureIgnoreCase) && envRec.EMail.Equals(recipEmail, StringComparison.CurrentCultureIgnoreCase)) {
                                orderIndex = envRec.OrderIndex;
                            }
                            r++;
                        }            
                        
                        string recipientOverrideXml = File.ReadAllText(@"..\..\..\Ressources\RecipientOverride.xml");
                        recipientOverrideXml = recipientOverrideXml.Replace("##recipLookupEmail##", recipEmail);
                        recipientOverrideXml = recipientOverrideXml.Replace("##orderIndex##", XmlConvert.ToString(orderIndex));

                        string replaceResponseXml = ws.ReplaceRecipient_v1(auth,envelopeList[i].InnerText,recipientOverrideXml);

                        SignAnywhereResponse replaceResponse = SignAnywhereResponse.parseResponse(replaceResponseXml);
                        if (replaceResponse.BaseResult != SignAnywhereResponse.Result.Ok)
                        {
                            Console.WriteLine(replaceResponse.Error + Environment.NewLine + replaceResponse.ErrorMsg);
                        }
                        else {
                            Console.WriteLine("Recipient replaced.");
                        }

                        // in case, the same recipient is used as CC recipient (or later in the envelope as signer again), 
                        // and you want to replace this recipient too, call ReplaceRecipient_v1 again with the appropriate order index
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }
    }

}
