﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace eSignAnyWhere_SampleCode_SOAP
{
    class UserManagement
    {
        private static eSAW.UserManagement.UserManagementApi GetUserManagementWebService()
        {
            eSAW.UserManagement.UserManagementApi ws = new eSAW.UserManagement.UserManagementApi();
            ws.Url = Program.eSawUrl + @"/UserManagementApi.asmx";

            // for onPremise version with WindowsAuthentication only: set user token
            ws.Credentials = System.Net.CredentialCache.DefaultCredentials;

            return ws;
        }


        public static void FindUsers()
        {
            try
            {
                eSAW.UserManagement.UserManagementApi ws = GetUserManagementWebService();
                string auth = Program.GetApiAuthorization();


                string userFindDescriptorXml = "<findUsersDescriptor />";
                //<findUsersDescriptor>
                //    <role>None|PowerUser|RegisteredSigner</role> <!-- optional -->
                //    <isUserManager>0|1|true|false</isUserManager> <!-- optional -->
                //</findUsersDescriptor>

                string findUsersResponse = ws.FindUsers_v2(auth, userFindDescriptorXml);               

                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(findUsersResponse);
                if (response.BaseResult == SignAnywhereResponse.Result.Ok)
                {
                    Console.WriteLine("List of found users:");
                    Console.WriteLine(response.okInfo.InnerXml);

                }
                else
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg);
                }

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }


        public static void CreateUser()
        {
            try
            {
                eSAW.UserManagement.UserManagementApi ws = GetUserManagementWebService();
                string auth = Program.GetApiAuthorization();


                string userDescriptionXml = @"
<userDescription>
	<eMail>adamtester@xyzmo.com</eMail>
	<firstName>Adam</firstName>
	<lastName>Tester</lastName>
	<role>PowerUser</role>
	<isUserManager>true</isUserManager>
	<notifyWhenActionNeeded>true</notifyWhenActionNeeded>
	<notifyWhenRecipientSigned>true</notifyWhenRecipientSigned>
	<notifyWhenRecipientRejected>true</notifyWhenRecipientRejected>
	<notifyWhenDeliveryFailed>true</notifyWhenDeliveryFailed>
	<notifyWhenRecipientOpenedDocument>true</notifyWhenRecipientOpenedDocument>
	<authentications >
		<samlAuthentication >
			<providerName >Internal AD FS</providerName >
			<userIdentifierAttibuteValue >adamtester@xyzmo.com</userIdentifierAttibuteValue >    
		</samlAuthentication >                 
	</authentications >
</userDescription>";
                
                //<userDescription>
                //    <eMail>some@email.com</eMail>
                //    <firstName>Jonn</firstName>
                //    <lastName>Doe</lastName>
                //    <authentications>
                //      <samlAuthentication> <!-- array -->>
                //        <providerName>AD FS My</providerName> <!-- SAML (AD FS, authentity provider with SAML) -->
                //        <userIdentifierAttibuteValue>adam</userIdentifierAttibuteValue> <!-- SAML  (AD FS) -->
                //      </samlAuthentication>
                //    </authentications>
                //    <jobTitle/><!-- optional -->
                //    <phoneNumber/><!-- optional -->
                //    <role>None|PowerUser|RegisteredSigner</role>
                //    <isUserManager>0|1|true|false</isUserManager>
                //    <defaultSubject/> <!-- optional -->
                //    <defaultMessage/><!-- optional -->
                //    <notifyWhenActionNeeded>0|1|true|false</notifyWhenActionNeeded>
                //    <notifyWhenRecipientSigned>0|1|true|false</notifyWhenRecipientSigned>
                //    <notifyWhenRecipientRejected>0|1|true|false</notifyWhenRecipientRejected>
                //    <notifyWhenDeliveryFailed>0|1|true|false</notifyWhenDeliveryFailed>
                //    <notifyWhenRecipientOpenedDocument>0|1|true|false</notifyWhenRecipientOpenedDocument>
                //</userDescription>

                string userActivationDescriptorXml = "<userActivationDescriptor><sendInvitationMail>false</sendInvitationMail></userActivationDescriptor>";

                //<userActivationDescriptor>
                //    <sendInvitationMail>0|1|true|false</sendInvitationMail>
                //</userActivationDescriptor>
               
                string createUserResponse = ws.CreateUser_v1(auth, userDescriptionXml, userActivationDescriptorXml);
               
                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(createUserResponse);
                if (response.BaseResult == SignAnywhereResponse.Result.Ok)
                {
                    Console.WriteLine("User sucessfully created ...");
                }
                else
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg);
                }

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }

        public static void UpdateUser() {

            try
            {
                eSAW.UserManagement.UserManagementApi ws = GetUserManagementWebService();
                string auth = Program.GetApiAuthorization();

                string getUserDetailsResponseStr = ws.GetUserByEmail_v1(auth, "some@email.com");

                SignAnywhereResponse userDetailsResponse = SignAnywhereResponse.parseResponse(getUserDetailsResponseStr);
                if (userDetailsResponse.BaseResult != SignAnywhereResponse.Result.Ok)
                {
                    Console.WriteLine(userDetailsResponse.Error + Environment.NewLine + userDetailsResponse.ErrorMsg);
                    return;
                    
                }

                string userId = userDetailsResponse.okInfo.SelectSingleNode("user/id").InnerText;

                string updateUserDescriptionXml = @"
<userDescription>	
	<firstName>Adam New Name</firstName>
</userDescription>";

                string updateUserResponse = ws.UpdateUser_v1(auth, userId, updateUserDescriptionXml);

                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(updateUserResponse);
                if (response.BaseResult == SignAnywhereResponse.Result.Ok)
                {
                    Console.WriteLine("User sucessfully updated ...");
                }
                else
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg);
                }
            }          
            
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }

         public static void DeleteUser()
        {
            try
            {
                eSAW.UserManagement.UserManagementApi ws = GetUserManagementWebService();
                string auth = Program.GetApiAuthorization();


                string reassignEnvelopesDescriptorXml = "<reassignEnvelopesDescriptor>  <newRecipient>newUser.mail@gmail.com</newRecipient>  <reassignDrafts>true</reassignDrafts>  <reassignTemplates>true</reassignTemplates>  <reassignClipboard>true</reassignClipboard>  <reassignAddressBook>true</reassignAddressBook></reassignEnvelopesDescriptor>";
                //<reassignEnvelopesDescriptor>
                //  <newRecipient>some@guy.com</newRecipient>
                //  <reassignDrafts>0|1|true|false</reassignDrafts>
                //  <reassignTemplates>0|1|true|false</reassignTemplates>
                //  <reassignClipboard>0|1|true|false</reassignClipboard>
                //  <reassignAddressBook>0|1|true|false</reassignAddressBook>
                //</reassignEnvelopesDescriptor>

                string deleteUserResponse = ws.DeleteUser_v1(auth, "some.mail@gmail.com", reassignEnvelopesDescriptorXml);

                SignAnywhereResponse response = SignAnywhereResponse.parseResponse(deleteUserResponse);
                if (response.BaseResult == SignAnywhereResponse.Result.Ok)
                {
                    Console.WriteLine("User deleted ...");            

                }
                else
                {
                    Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg);
                }

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }   
        }



         public static void GetTeamsAndSetTeams() {

             try
             {
                 eSAW.UserManagement.UserManagementApi ws = GetUserManagementWebService();
                 string auth = Program.GetApiAuthorization();




                 string getTeamResponse = ws.GetTeams_v1(auth);

                 SignAnywhereResponse response = SignAnywhereResponse.parseResponse(getTeamResponse);
                 if (response.BaseResult == SignAnywhereResponse.Result.Ok)
                 {
                     Console.WriteLine(response.okInfo.InnerXml);
                      // structure of result has to be used as input to define the new teamstructure in the organisatin

                      // <teams>
                      //   <team>
                      //      <name>test team</name>
                      //      <allowEnvelopeSharingWithinTeam>false</allowEnvelopeSharingWithinTeam>
                      //      <allowTemplateSharingWithinTeam>false</allowTemplateSharingWithinTeam>
                      //      <teamMember>
                      //         <eMail>knorze@xyzmo.com</eMail>
                      //         <teamMembers>
                      //            <teamMember>
                      //               <eMail>ronald.korze@significant.com</eMail>
                      //               <teamMembers />
                      //            </teamMember>
                      //         </teamMembers>
                      //      </teamMember>
                      //   </team>
                      //</teams>

                     string setTeamResponse = ws.SetTeams_v1(auth, response.okInfo.InnerXml);

                     SignAnywhereResponse response2 = SignAnywhereResponse.parseResponse(setTeamResponse);
                     if (response2.BaseResult == SignAnywhereResponse.Result.Ok)
                     {
                         Console.WriteLine("New team structure successfully set ...");

                     }
                     else
                     {
                         Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg);
                     }
                 }
                 else
                 {
                     Console.WriteLine(response.Error + Environment.NewLine + response.ErrorMsg);
                 }
             }
             catch (Exception ex)
             {
                 Console.WriteLine(ex.ToString());
             }         
         
         
         }
    }
}
