﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Microsoft.Rest;
using Microsoft.Rest.Serialization;
using System.Net;
using System.IO;

using eSaw.Models;
using eSaw;

namespace eSignAnyWhere_SampleCode_REST
{
    class Program
    {

        //http://beta4.testlab.xyzmo.com/api/swagger/docs/V4.0
        public static eSaw.ESawClient eSawAPI;

        public const string eSawUrl = "http://beta4.testlab.xyzmo.com";
        const string UserAccountName = "your@email.com";
        const string OrganizationKey = "48024f64-c2b4-403a-ac16-80c287b2fa2d";  //they value in the eSAW UI: Settings/Organization/OrganizationDetails -> Organization Key

        // finally, the url is called: <serverUrl>/api/v2.0/version" 
        // the version is added via generated code, however, in your integration make sure to always 
        const string eSawUrlAPI = eSawUrl + "/api";



        static void Main(string[] args)

        {
            RunAsync().GetAwaiter().GetResult();
        }

        /* Please note, that eSignAnywhere consist in reality of 2 components: 
        *                a.) The workflow management system eSignAnywhere  - and - 
        *                b.) The system that takes care about each single step in the workflow: Workstep Controller and SAW Viewer (the front end for signing the document). Both are part of SIGNificant Server Platform (SSP)
        *                
        * The sample code shown below focuses on the eSignAnywhere specific part of the system. 
        * Whenever if comes to a specific configuration of a single step in the workflow, or for a dedicated behavior of Saw Viewer, that you want to control, please have a look here as well:
        *       
        *      https://partner.xyzmo.com/projects/significant-server-platform/wiki/WSC_HowTo_Scenarios
        *      https://partner.xyzmo.com/projects/significant-server-platform/wiki/WSC_HowTo_Scenarios#Full-WorkstepConfiguration
        *      https://partner.xyzmo.com/projects/significant-server-platform/wiki/SAW_Integration_Guide#ViewerPreferences
        * 
        * In case you can't access these resources, please get into contact with your presales contact or contact our support.
        * 
        */

        static async Task RunAsync()
        {

            eSawAPI = new eSaw.ESawClient(new Uri(eSawUrlAPI), new Microsoft.Rest.BasicAuthenticationCredentials());

            await GetVersion();

            await ValidateAuthorization();

            //await SendEnvelopeSimple();
            //await SendEnvelopeSimple_AddSignatureFields_Biometric_And_DisposableCertificate_and_AddFormFields_AddDocumentTimestamp();

            //await SendEnvelopeWith2SignersAndCopyReceiverAndPrefillFormFieldValues();
            //await SendEnvelope_POS_Scenario();

            //await CreateDraft(new string[] { @"..\..\..\Ressources\Demo_Contract_Form.pdf", @"..\..\..\Ressources\Title.pdf" });

            //await SendEnvelopeFromTemplate("a217eba4-2582-4d9e-a369-4de4f8840aa2");
            //await CreateDraftFromTemplate("355f88f9-3de4-499a-a0a0-35b15a9ae1e9");

            //await GetEnvelopeStatus("22e4e31f-90bf-48a0-8338-27b32d5e9a91");

            //await DownloadCompletedEnvelope("22e4e31f-90bf-48a0-8338-27b32d5e9a91");

            //await DeleteEnvelope("22e4e31f-90bf-48a0-8338-27b32d5e9a91");

            //await FindEnvelopes();
            //await FindAllEnvelopsThatAreWaitingForDedicatedRecipientAndReplaceRecipient();

            //await FindUsers();

            Console.Read();

        }

        static async Task GetVersion()
        {
            try
            {
                var result = await eSawAPI.Version.GetWithHttpMessagesAsync();
                Console.WriteLine("Version: " + result.Body.Version);
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }

        static Dictionary<string, List<string>> HelperGetAuthorizationHeader()
        {
            Dictionary<string, List<string>> auth = new Dictionary<string, List<string>> { { "organizationKey", new List<string> { OrganizationKey } },
                { "userLoginName", new  List<string> { UserAccountName } } };
            return auth;
        }

        static async Task ValidateAuthorization()
        {
            try
            {
                var result = await eSawAPI.Authorization.ValidateWithHttpMessagesAsync(HelperGetAuthorizationHeader());
                Console.WriteLine("Authorization Valid: " + result.Response.IsSuccessStatusCode);
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }



        static async Task SendEnvelopeSimple()
        {
            try
            {
                FileStream fileToUpload = File.Open(@"..\..\..\Ressources\Demo_Contract_Form.pdf", FileMode.Open);

                var result = await eSawAPI.SspFile.UploadTemporaryWithHttpMessagesAsync(fileToUpload, HelperGetAuthorizationHeader());
                string docId = "";
                if (result.Response.IsSuccessStatusCode)
                {
                    eSaw.Models.UploadSspFileResult uploadResultSuccessObj = result.Body as eSaw.Models.UploadSspFileResult;
                    docId = uploadResultSuccessObj.SspFileId;
                }
                else
                {
                    eSaw.Models.Error flowError = result.Body as eSaw.Models.Error;
                    Console.WriteLine("Upload document failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                    return;
                }

                eSaw.Models.WorkstepConfiguration workstepConfig = null;
                List<string> sspFileIds = new List<string> { docId };

                eSaw.Models.EnvelopePrepareModel prepareConfig = new eSaw.Models.EnvelopePrepareModel(); // Attention: always use the empty constructor (on all api proxy classes)
                //the empty constructor will simplify updates to newer versions of the interface
                //the constructor that allows you to assign properties directly can result in a silent error on changed order of properties or new properties
                prepareConfig.SspFileIds = sspFileIds;
                prepareConfig.PrepareSendEnvelopeStepsDescriptor = new eSaw.Models.PrepareSendEnvelopeStepsDescriptor();
                prepareConfig.PrepareSendEnvelopeStepsDescriptor.ClearFieldMarkupString = false;
                prepareConfig.AdHocWorkstepConfiguration = new eSaw.Models.AdHocWorkstepConfiguration();
                // define the list of available signature types per signature. This can be modified later on as well per each signature, according to your needs, just before Envelope.Send
                prepareConfig.AdHocWorkstepConfiguration.SigTemplate = new SigTemplate();
                prepareConfig.AdHocWorkstepConfiguration.SigTemplate.AllowedSignatureTypes = new List<SigType>();
                SigTypeClick2Sign sigTypeClick2Sign = new SigTypeClick2Sign();
                sigTypeClick2Sign.Id = "1";
                prepareConfig.AdHocWorkstepConfiguration.SigTemplate.AllowedSignatureTypes.Add(sigTypeClick2Sign);

                var prepareResult = await eSawAPI.Envelope.PrepareWithHttpMessagesAsync(prepareConfig, HelperGetAuthorizationHeader());
                if (prepareResult.Response.IsSuccessStatusCode)
                {

                    PrepareSendEnvelopeStepsResult prepareResultSuccess = prepareResult.Body as PrepareSendEnvelopeStepsResult;
                    workstepConfig = prepareResultSuccess.AdHocWorkstepConfigResult;
                }
                else
                {
                    eSaw.Models.Error flowError = prepareResult.Body as Error;
                    Console.WriteLine("Prepare envelope failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                    return;
                }

                eSaw.Models.EnvelopeSendModel envelopeConfig = new eSaw.Models.EnvelopeSendModel();
                envelopeConfig.SspFileIds = sspFileIds;
                envelopeConfig.SendEnvelopeDescription = new eSaw.Models.SendEnvelopeDescription();
                envelopeConfig.SendEnvelopeDescription.Name = "Envelope via Rest Api " + DateTime.Now.ToLongDateString() + " " + DateTime.Now.ToLongTimeString();
                envelopeConfig.SendEnvelopeDescription.EmailSubject = "my envelope via rest ";
                envelopeConfig.SendEnvelopeDescription.EmailBody = "text for you ...";                

                envelopeConfig.SendEnvelopeDescription.Steps = new List<eSaw.Models.SendEnvelopeStep>();

                eSaw.Models.SendEnvelopeStep firstStep = new eSaw.Models.SendEnvelopeStep();
                envelopeConfig.SendEnvelopeDescription.Steps.Add(firstStep);
                firstStep.WorkstepConfiguration = workstepConfig;
                firstStep.Recipients = new List<eSaw.Models.SendEnvelopeRecipient> { new eSaw.Models.SendEnvelopeRecipient() };
                firstStep.Recipients[0].Email = "first.signer@email.com";
                firstStep.Recipients[0].LastName = "Signer";
                firstStep.Recipients[0].FirstName = "First";
                firstStep.Recipients[0].LanguageCode = "en";

                var sendEnvelopeResult = await eSawAPI.Envelope.SendWithHttpMessagesAsync(envelopeConfig, HelperGetAuthorizationHeader());

                if (sendEnvelopeResult.Response.IsSuccessStatusCode)
                {
                    eSaw.Models.SendEnvelopeResult sendEnvelopeResultSuccessObj = sendEnvelopeResult.Body as SendEnvelopeResult;
                    string envelopeId = sendEnvelopeResultSuccessObj.EnvelopeId;
                    Console.WriteLine("Send Envelope successful: " + envelopeId);
                }
                else
                {
                    eSaw.Models.Error flowError = sendEnvelopeResult.Body as Error;
                    Console.WriteLine("Send envelope failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }

        static async Task SendEnvelopeSimple_AddSignatureFields_Biometric_And_DisposableCertificate_and_AddFormFields_AddDocumentTimestamp()
        {
            try
            {
                FileStream fileToUpload = File.Open(@"..\..\..\Ressources\Title.pdf", FileMode.Open);

                var result = await eSawAPI.SspFile.UploadTemporaryWithHttpMessagesAsync(fileToUpload, HelperGetAuthorizationHeader());
                string docId = "";
                if (result.Response.IsSuccessStatusCode)
                {
                    eSaw.Models.UploadSspFileResult uploadResultSuccessObj = result.Body as eSaw.Models.UploadSspFileResult;
                    docId = uploadResultSuccessObj.SspFileId;
                }
                else
                {
                    eSaw.Models.Error flowError = result.Body as eSaw.Models.Error;
                    Console.WriteLine("Upload document failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                    return;
                }


                List<string> sspFileIds = new List<string> { docId };

                // pdf is empty and does not contain signature fields, testmarkers for signature fields, no form fields
                // therefore no Prepare call is necessary

                eSaw.Models.EnvelopeSendModel envelopeConfig = new eSaw.Models.EnvelopeSendModel();// Attention: always use the empty constructor (on all api proxy classes)
                //the empty constructor will simplify updates to newer versions of the interface
                //the constructor that allows you to assign properties directly can result in a silent error on changed order of properties or new properties
                envelopeConfig.SspFileIds = sspFileIds;
                envelopeConfig.SendEnvelopeDescription = new eSaw.Models.SendEnvelopeDescription();
                envelopeConfig.SendEnvelopeDescription.Name = "Envelope via Rest Api Additional Signature Fields" + DateTime.Now.ToLongDateString() + " " + DateTime.Now.ToLongTimeString();
                envelopeConfig.SendEnvelopeDescription.EmailSubject = "my envelope via rest ";
                envelopeConfig.SendEnvelopeDescription.EmailBody = "text for you ...";
                

                // adding form fields to the pdf         
                #region adding form fields to pdf
                envelopeConfig.SendEnvelopeDescription.AddFormFields = new AddFormFields();
                envelopeConfig.SendEnvelopeDescription.AddFormFields.Forms = new Dictionary<string, IList<AddFormFieldDescriptor>>();


                IList<AddFormFieldDescriptor> formFieldsToAddForDocument1 = new List<AddFormFieldDescriptor>();
                envelopeConfig.SendEnvelopeDescription.AddFormFields.Forms.Add("1", formFieldsToAddForDocument1); //docIndex as key

                AddTextBoxFormFieldDescriptor tbToAdd1 = new AddTextBoxFormFieldDescriptor();
                formFieldsToAddForDocument1.Add(tbToAdd1);                
                tbToAdd1.Name = "Location";
                tbToAdd1.Page = 1;
                tbToAdd1.X = 100;
                tbToAdd1.Y = 546;
                tbToAdd1.Width = 200;
                tbToAdd1.Height = 24;
                tbToAdd1.FontSize = 12;
                tbToAdd1.FontName = "Arial";
                tbToAdd1.Value = "MyLocation";                 

                AddCheckBoxFormFieldDescriptor cbToAdd1 = new AddCheckBoxFormFieldDescriptor();
                formFieldsToAddForDocument1.Add(cbToAdd1);                
                cbToAdd1.Name = "cbAgreement";
                cbToAdd1.Page = 1;
                cbToAdd1.X = 360;
                cbToAdd1.Y = 546;
                cbToAdd1.Width = 24;
                cbToAdd1.Height = 24;
                cbToAdd1.IsChecked = true;
                cbToAdd1.ExportValue = "YES";
                #endregion
                //
                envelopeConfig.SendEnvelopeDescription.Steps = new List<eSaw.Models.SendEnvelopeStep>();

                eSaw.Models.SendEnvelopeStep firstStep = new eSaw.Models.SendEnvelopeStep();
                envelopeConfig.SendEnvelopeDescription.Steps.Add(firstStep);

                firstStep.Recipients = new List<eSaw.Models.SendEnvelopeRecipient> { new eSaw.Models.SendEnvelopeRecipient() };
                firstStep.Recipients[0].AllowDelegation = true;

                firstStep.Recipients[0].Email = "first@signer.com";
                firstStep.Recipients[0].LastName = "first";
                firstStep.Recipients[0].FirstName = "signer";
                firstStep.Recipients[0].LanguageCode = "en";

                //additional data for disposable certficate
                DisposableCertificateData recip1DisposableCertData = new DisposableCertificateData();
                firstStep.Recipients[0].DisposableCertificateData = recip1DisposableCertData;

                recip1DisposableCertData.PhoneMobile = "+4366412345678";
                recip1DisposableCertData.CountryResidence = "AT"; //iso country code
                recip1DisposableCertData.DocumentType =  DisposableCertificateDocumentTypeEnum.PASS;
                recip1DisposableCertData.DocumentNumber = "P3912356";
                recip1DisposableCertData.DocumentIssuedOn = new DateTime(2009, 10, 16);
                recip1DisposableCertData.DocumentExpiryDate = new DateTime(2019, 10, 15);
                recip1DisposableCertData.DocumentIssuedBy = "Some Authority";
                recip1DisposableCertData.IdentificationCountry = "AT";
                recip1DisposableCertData.IdentificationType = DisposableCertificateIdentificationTypeEnum.FOREIGNTAXCODE;
                recip1DisposableCertData.SerialNumber = "200012365";

                //                

                //define workstep config for recipient
                eSaw.Models.WorkstepConfiguration workstepConfig = new WorkstepConfiguration();
                firstStep.WorkstepConfiguration = workstepConfig;

                workstepConfig.Policy = new Policy();
                workstepConfig.Policy.WorkstepTasks = new WorkstepTasks();
                workstepConfig.Policy.GeneralPolicies = new GeneralPolicies();
                workstepConfig.Policy.GeneralPolicies.AllowRejectWorkstep = true;
                workstepConfig.Policy.WorkstepTasks.Tasks = new List<WorkstepTask>();
                
                

                //workstepConfig.Policy.WorkstepTasks.ReferenceCorner = ReferenceCornerType.LowerLeft;

                //let's define a form filling task
                FormGroup fillFormTask = new FormGroup();
                workstepConfig.Policy.WorkstepTasks.Tasks.Add(fillFormTask);                
                fillFormTask.Id = "FillForm1";
                fillFormTask.DisplayName = "Fill Form Fields";
                fillFormTask.DocRefNumber = 1;
                fillFormTask.Forms = new List<Form>();

                TextBox tb1 = new TextBox();                
                fillFormTask.Forms.Add(tb1);
                tb1.Id = "Location";
                tb1.DocRefNumber = 1;
                //tb1.KeepExistingValue = true;
                

                CheckBox cb1 = new CheckBox();                
                fillFormTask.Forms.Add(cb1);
                cb1.DocRefNumber = 1;
                cb1.Id = "cbAgreement";
                //cb1.KeepExistingValue = true;                

                //now let's add some signature fields
                Signature sigField1 = new Signature();
                //sigField1.TaskType = WorkstepTaskType.SignField; 
                sigField1.Id = "sigField1"; //make sure the id is unique per document (over all signers!!)
                sigField1.DisplayName = "Sign Signature Field 1";
                sigField1.PositionPage = 1;
                sigField1.DocRefNumber = 1;
                sigField1.Position = new Position(100, 440);
                sigField1.Size = new Size(80, 160);
                sigField1.IsRequired = true;

                //optional, if you want to specify one or more certain signature types
                //default: biometric signature
                //note: some signature types (remote signature, sms-otp, disposable remote certificate) require additional data specified for the signer (see above)
                sigField1.AllowedSignatureTypes = new List<SigType>();
                SigTypeBiometricSignature sigField1BiometricSigType = new SigTypeBiometricSignature();                
                sigField1.AllowedSignatureTypes.Add(sigField1BiometricSigType);               
                sigField1BiometricSigType.Id = "1";
                sigField1BiometricSigType.Preferred = true;

                SigTypePicture sigField1Draw2Sign = new SigTypeDraw2Sign();                
                sigField1Draw2Sign.Id = "2";
                sigField1.AllowedSignatureTypes.Add(sigField1Draw2Sign);
                workstepConfig.Policy.WorkstepTasks.Tasks.Add(sigField1); //add to task list

                // add signature field for disposable certificate
                Signature sigField2 = new Signature();
                workstepConfig.Policy.WorkstepTasks.Tasks.Add(sigField2);               
                sigField2.Id = "sigField2";
                sigField2.DisplayName = "Sign Signature Field 2";
                sigField2.DocRefNumber = 1;
                sigField2.PositionPage = 1;
                sigField2.Position = new Position(360, 440);
                sigField2.Size = new Size(80, 160);
                //sigField2.IsRequired = true;

                sigField2.AllowedSignatureTypes = new List<SigType>();
                SigTypeTransactionCode transCodeSig = new SigTypeTransactionCode();
                transCodeSig.Id = "1";                
                transCodeSig.TrModType = "RemoteSignatureDisposableCertificate";
                transCodeSig.TrValidityInSeconds = 300;
                transCodeSig.Ly = "remoteCertificate";
                transCodeSig.TrConfId = "remoteCertificateSignSmsText";
                transCodeSig.TrConfIdIssueCertificate = "disposableCertificateEnrolAndSignSmsText";
                sigField2.AllowedSignatureTypes.Add(transCodeSig);

                // add document timestamp 
                workstepConfig.Policy.FinalizeActions = new FinalizeActions();                
                workstepConfig.Policy.FinalizeActions.FinalizeActionList = new List<FinalizeAction>();
                Timestamp documentTimestamp = new Timestamp();
                documentTimestamp.DocRefNumbers = "*";
                workstepConfig.Policy.FinalizeActions.FinalizeActionList.Add(documentTimestamp);

                // end of defining workstep config                
                var sendEnvelopeResult = await eSawAPI.Envelope.SendWithHttpMessagesAsync(envelopeConfig, HelperGetAuthorizationHeader());

                if (sendEnvelopeResult.Response.IsSuccessStatusCode)
                {
                    eSaw.Models.SendEnvelopeResult sendEnvelopeResultSuccessObj = sendEnvelopeResult.Body as SendEnvelopeResult;
                    string envelopeId = sendEnvelopeResultSuccessObj.EnvelopeId;
                    Console.WriteLine("Send Envelope successful: " + envelopeId);
                }
                else
                {
                    eSaw.Models.Error flowError = sendEnvelopeResult.Body as Error;
                    Console.WriteLine("Send envelope failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }

        static async Task SendEnvelopeWith2SignersAndCopyReceiverAndPrefillFormFieldValues()
        {
            try
            {
                FileStream fileToUpload = File.OpenRead(@"..\..\..\Ressources\Demo_Contract_Form.pdf");

                var result = await eSawAPI.SspFile.UploadTemporaryWithHttpMessagesAsync(fileToUpload, HelperGetAuthorizationHeader());
                string docId = "";
                if (result.Response.IsSuccessStatusCode)
                {
                    eSaw.Models.UploadSspFileResult uploadResultSuccessObj = result.Body as eSaw.Models.UploadSspFileResult;
                    docId = uploadResultSuccessObj.SspFileId;
                }
                else
                {
                    eSaw.Models.Error flowError = result.Body as eSaw.Models.Error;
                    Console.WriteLine("Upload document failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                    return;
                }

                eSaw.Envelope eSawApiEnvelopeInstance = new Envelope(eSawAPI);

                eSaw.Models.WorkstepConfiguration workstepConfig1stSigner = null;
                eSaw.Models.WorkstepConfiguration workstepConfig2ndSigner = null;

                List<string> sspFileIds = new List<string> { docId };

                eSaw.Models.EnvelopePrepareModel prepareConfig = new eSaw.Models.EnvelopePrepareModel();
                prepareConfig.SspFileIds = sspFileIds;
                prepareConfig.PrepareSendEnvelopeStepsDescriptor = new eSaw.Models.PrepareSendEnvelopeStepsDescriptor();
                prepareConfig.PrepareSendEnvelopeStepsDescriptor.ClearFieldMarkupString = false;
                prepareConfig.AdHocWorkstepConfiguration = new eSaw.Models.AdHocWorkstepConfiguration();

                var prepareResult = await eSawApiEnvelopeInstance.PrepareWithHttpMessagesAsync(prepareConfig, HelperGetAuthorizationHeader());
                if (prepareResult.Response.IsSuccessStatusCode)
                {

                    PrepareSendEnvelopeStepsResult prepareEnvelopeResultSuccess1stSigner = prepareResult.Body as PrepareSendEnvelopeStepsResult;
                    workstepConfig1stSigner = prepareEnvelopeResultSuccess1stSigner.AdHocWorkstepConfigResult;

                    // !!! IMPORTANT !!!
                    // get additional (independent object) for second signer ... 
                    // therefore I simply deserialize the response again (a deep copy of the workstep config would be fine as well)
                    string responseContent2ndSigner = await prepareResult.Response.Content.ReadAsStringAsync().ConfigureAwait(false);
                    PrepareSendEnvelopeStepsResult sendEnvelopeResultSuccess2ndSigner = SafeJsonConvert.DeserializeObject<PrepareSendEnvelopeStepsResult>(responseContent2ndSigner, eSawApiEnvelopeInstance.Client.DeserializationSettings);
                    workstepConfig2ndSigner = sendEnvelopeResultSuccess2ndSigner.AdHocWorkstepConfigResult;
                }
                else
                {
                    eSaw.Models.Error flowError = prepareResult.Body as Error;
                    Console.WriteLine("Prepare envelope failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                    return;
                }

                // compose envelope config
                eSaw.Models.EnvelopeSendModel envelopeConfig = new eSaw.Models.EnvelopeSendModel();
                envelopeConfig.SspFileIds = sspFileIds;
                envelopeConfig.SendEnvelopeDescription = new eSaw.Models.SendEnvelopeDescription();
                envelopeConfig.SendEnvelopeDescription.Name = "Envelope via Rest Api " + DateTime.Now.ToLongDateString() + " " + DateTime.Now.ToLongTimeString();
                envelopeConfig.SendEnvelopeDescription.EmailSubject = "my envelope via rest ";
                envelopeConfig.SendEnvelopeDescription.EmailBody = "text for you ...";               

                // override values of some form fields
                envelopeConfig.SendEnvelopeDescription.OverrideFormFieldValues = new OverrideFormFieldOptions();
                envelopeConfig.SendEnvelopeDescription.OverrideFormFieldValues.Forms = new Dictionary<string, IList<FillFormDescriptor>>();

                List<eSaw.Models.FillFormDescriptor> newFormFieldValuesForDoc1 = new List<FillFormDescriptor>();
                envelopeConfig.SendEnvelopeDescription.OverrideFormFieldValues.Forms.Add("1", newFormFieldValuesForDoc1);

                FillTextBoxFormDescriptor tb1 = new FillTextBoxFormDescriptor();
                tb1.Name = "Text1";
                tb1.Value = "My new value in text1";
                newFormFieldValuesForDoc1.Add(tb1);

                FillTextBoxFormDescriptor tb2 = new FillTextBoxFormDescriptor();
                tb2.Name = "Text2";
                tb2.Value = "sample value in text2";
                newFormFieldValuesForDoc1.Add(tb2);

                FillRadioButtonFormDescriptor rbGroup3 = new FillRadioButtonFormDescriptor();
                rbGroup3.Name = "Group3";
                rbGroup3.SelectedItem= "Group3#XyzmoIndexSeparator#0";
                newFormFieldValuesForDoc1.Add(rbGroup3);

                FillCheckBoxFormDescriptor cbClientAgreesPayDown = new FillCheckBoxFormDescriptor();
                cbClientAgreesPayDown.Name = "CheckBox_ClientAgreesPayDown";
                cbClientAgreesPayDown.IsChecked = true;
                newFormFieldValuesForDoc1.Add(cbClientAgreesPayDown);

                envelopeConfig.SendEnvelopeDescription.Steps = new List<eSaw.Models.SendEnvelopeStep>();

                // add step for first signer

                // adjust workstep config according to your needs for first signer
                // in the demo form, I remove the last form field on page 4 and I remove the 2nd signature field (those should be managed by the second signer)

                eSaw.Models.FormGroup formFillingTask3ForSigner1 = (FormGroup)workstepConfig1stSigner.Policy.WorkstepTasks.Tasks[2];
                formFillingTask3ForSigner1.Forms.RemoveAt(3);
                workstepConfig1stSigner.Policy.WorkstepTasks.Tasks.RemoveAt(4);

                //
                eSaw.Models.SendEnvelopeStep firstStep = new eSaw.Models.SendEnvelopeStep();
                envelopeConfig.SendEnvelopeDescription.Steps.Add(firstStep);
                firstStep.RecipientType = RecipientTypeEnum.Signer; 
                firstStep.OrderIndex = 0;
                firstStep.WorkstepConfiguration = workstepConfig1stSigner;
                firstStep.Recipients = new List<eSaw.Models.SendEnvelopeRecipient> { new eSaw.Models.SendEnvelopeRecipient() };
                firstStep.Recipients[0].Email = "first.signer@email.com";
                firstStep.Recipients[0].LastName = "Signer";
                firstStep.Recipients[0].FirstName = "First";
                firstStep.Recipients[0].LanguageCode = "en";

                // add step for 2nd signer

                // adjust workstep config according to your needs for 2nd signer
                // I remove now all the fields, that were assigned to the first signer already
                workstepConfig2ndSigner.Policy.WorkstepTasks.Tasks.RemoveAt(0);
                workstepConfig2ndSigner.Policy.WorkstepTasks.Tasks.RemoveAt(0);
                eSaw.Models.FormGroup formFillingTask3ForSigner2 = (FormGroup)workstepConfig2ndSigner.Policy.WorkstepTasks.Tasks[0];
                Form formFieldForSigner2 = formFillingTask3ForSigner2.Forms[3];
                formFillingTask3ForSigner2.Forms.Clear();
                formFillingTask3ForSigner2.Forms.Add(formFieldForSigner2);
                workstepConfig2ndSigner.Policy.WorkstepTasks.Tasks.RemoveAt(1);

                eSaw.Models.SendEnvelopeStep secondStep = new eSaw.Models.SendEnvelopeStep();
                envelopeConfig.SendEnvelopeDescription.Steps.Add(secondStep);
                secondStep.WorkstepConfiguration = workstepConfig2ndSigner;
                secondStep.RecipientType = RecipientTypeEnum.Signer;
                secondStep.OrderIndex = 1;
                secondStep.Recipients = new List<eSaw.Models.SendEnvelopeRecipient> { new eSaw.Models.SendEnvelopeRecipient() };
                secondStep.Recipients[0].Email = "second.signer@email.com";
                secondStep.Recipients[0].LastName = "Signer";
                secondStep.Recipients[0].FirstName = "Second";
                secondStep.Recipients[0].LanguageCode = "en";

                // add copy receiver step
                eSaw.Models.SendEnvelopeStep thirdStepCopyReceiver = new eSaw.Models.SendEnvelopeStep();
                envelopeConfig.SendEnvelopeDescription.Steps.Add(thirdStepCopyReceiver);
                thirdStepCopyReceiver.RecipientType =  RecipientTypeEnum.Acknowledge;
                thirdStepCopyReceiver.OrderIndex = 2;
                thirdStepCopyReceiver.Recipients = new List<eSaw.Models.SendEnvelopeRecipient> { new eSaw.Models.SendEnvelopeRecipient() };
                thirdStepCopyReceiver.Recipients[0].Email = "ack@email.com";
                thirdStepCopyReceiver.Recipients[0].LastName = "Recipient";
                thirdStepCopyReceiver.Recipients[0].FirstName = "Ack";
                thirdStepCopyReceiver.Recipients[0].LanguageCode = "en";


                var sendEnvelopeResult = await eSawAPI.Envelope.SendWithHttpMessagesAsync(envelopeConfig, HelperGetAuthorizationHeader());

                if (sendEnvelopeResult.Response.IsSuccessStatusCode)
                {
                    eSaw.Models.SendEnvelopeResult sendEnvelopeResultSuccessObj = sendEnvelopeResult.Body as SendEnvelopeResult;
                    string envelopeId = sendEnvelopeResultSuccessObj.EnvelopeId;
                    Console.WriteLine("Send Envelope successful: " + envelopeId);
                }
                else
                {
                    eSaw.Models.Error flowError = sendEnvelopeResult.Body as Error;
                    Console.WriteLine("Send envelope failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }

        // this sample shows, how you can start an envelope within a POS scenario. The first step will be in a POS scenario, while other steps are as usual with notification to the user via email.
        // in a POS scenario it is not necessary to notify the customer, as he is directly in the branch - therefore 
        //    a) disable notification for this step
        //    b) directly open the SAW Viewer or (on premise only) choose one of the other SSP front ends (WSI, Kiosk SDK for SSP, mobile Apps (iOS, Android, Windows Phone/Tablet) for SSP)

        // for simplicity (to keep the code shorter) of the sample, only an envelope with 1 recipient is used, but additional recipients can be added, as demonstrated in sample "SendEnvelope_with2Recipients()"
        // remark: POS scenario step can't be a parallel steps (therefore only one step with order index 1 is allowed)
        static async Task SendEnvelope_POS_Scenario()
        {
            try
            {
                //as usual, upload file(s)
                FileStream fileToUpload = File.Open(@"..\..\..\Ressources\Demo_Contract_Form.pdf", FileMode.Open);

                var result = await eSawAPI.SspFile.UploadTemporaryWithHttpMessagesAsync(fileToUpload, HelperGetAuthorizationHeader());
                string docId = "";
                if (result.Response.IsSuccessStatusCode)
                {
                    eSaw.Models.UploadSspFileResult uploadResultSuccessObj = result.Body as eSaw.Models.UploadSspFileResult;
                    docId = uploadResultSuccessObj.SspFileId;
                }
                else
                {
                    eSaw.Models.Error flowError = result.Body as eSaw.Models.Error;
                    Console.WriteLine("Upload document failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                    return;
                }

                //as usual, compose envelopeConfig directly, or get help on this on using Envelope.Prepare
                eSaw.Models.WorkstepConfiguration workstepConfig = null;
                List<string> sspFileIds = new List<string> { docId };

                eSaw.Models.EnvelopePrepareModel prepareConfig = new eSaw.Models.EnvelopePrepareModel();
                prepareConfig.SspFileIds = sspFileIds;

                var prepareResult = await eSawAPI.Envelope.PrepareWithHttpMessagesAsync(prepareConfig, HelperGetAuthorizationHeader());
                if (prepareResult.Response.IsSuccessStatusCode)
                {
                    PrepareSendEnvelopeStepsResult sendEnvelopeResultSuccess = prepareResult.Body as PrepareSendEnvelopeStepsResult;
                    workstepConfig = sendEnvelopeResultSuccess.AdHocWorkstepConfigResult;
                }
                else
                {
                    eSaw.Models.Error flowError = prepareResult.Body as Error;
                    Console.WriteLine("Prepare envelope failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                    return;
                }

                eSaw.Models.EnvelopeSendModel envelopeConfig = new eSaw.Models.EnvelopeSendModel();
                envelopeConfig.SspFileIds = sspFileIds;
                envelopeConfig.SendEnvelopeDescription = new eSaw.Models.SendEnvelopeDescription();
                envelopeConfig.SendEnvelopeDescription.Name = "Envelope via Rest Api for POS" + DateTime.Now.ToLongDateString() + " " + DateTime.Now.ToLongTimeString();
                envelopeConfig.SendEnvelopeDescription.EmailSubject = "my email subject for the next steps";
                envelopeConfig.SendEnvelopeDescription.EmailBody = "my email body for the next steps";
                //TODO remove
                envelopeConfig.SendEnvelopeDescription.DaysUntilExpire = 28;

                envelopeConfig.SendEnvelopeDescription.Steps = new List<eSaw.Models.SendEnvelopeStep>();

                // a) disable email sending for recipient in first step
                eSaw.Models.SendEnvelopeStep firstStep = new eSaw.Models.SendEnvelopeStep();
                envelopeConfig.SendEnvelopeDescription.Steps.Add(firstStep);
                firstStep.WorkstepConfiguration = workstepConfig;
                firstStep.Recipients = new List<eSaw.Models.SendEnvelopeRecipient> { new eSaw.Models.SendEnvelopeRecipient() };
                firstStep.Recipients[0].DisableEmail = true; //no need to send out email, as POS scenario
                ///firstStep.Recipients[0].Email = "first.signer@email.com";// no need to specify email, but can be done as well
                firstStep.Recipients[0].LastName = "Signer";
                firstStep.Recipients[0].FirstName = "First";
                firstStep.Recipients[0].LanguageCode = "en";

                // as usual, send the envelope
                var sendEnvelopeResult = await eSawAPI.Envelope.SendWithHttpMessagesAsync(envelopeConfig, HelperGetAuthorizationHeader());
                string envelopeId = "";
                if (sendEnvelopeResult.Response.IsSuccessStatusCode)
                {
                    eSaw.Models.SendEnvelopeResult sendEnvelopeResultSuccessObj = sendEnvelopeResult.Body as SendEnvelopeResult;
                    envelopeId = sendEnvelopeResultSuccessObj.EnvelopeId;
                    Console.WriteLine("Send Envelope successful: " + envelopeId);
                }
                else
                {
                    eSaw.Models.Error flowError = sendEnvelopeResult.Body as Error;
                    Console.WriteLine("Send envelope failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                    return;
                }

                // b) open first step of envelope directly in your POS Application
                //     ba) get redirection url
                //     bb) only in case if other front end then SAW viewer should be used (on premise)
                //       bba) get workstep id
                //       bbb) compose url for other front end
                //     bc) open url (either for SAW Viewer or other front end)

                // ba) get redirection url

                string redirectionUrlFirstStep = "";
                var envelopeStatusResult = await eSawAPI.Envelope.ByIdWithHttpMessagesAsync(envelopeId, HelperGetAuthorizationHeader());

                if (envelopeStatusResult.Response.IsSuccessStatusCode)
                {
                    eSaw.Models.EnvelopeStatus envelopeStatusObj = envelopeStatusResult.Body as EnvelopeStatus;
                    redirectionUrlFirstStep = envelopeStatusObj.Bulks[0].Steps[0].WorkstepRedirectionUrl;
                }
                else
                {
                    eSaw.Models.Error flowError = envelopeStatusResult.Body as Error;
                    Console.WriteLine("Get Envelope Status failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                    return;
                }

                #region bb) optional: to use another front end
                // bb) optional: to use another front end
                // bba) get workstepId
                string requestUrlForWorkstepId = redirectionUrlFirstStep + "&responseType=returnWorkstepId";
                HttpWebRequest webRequest = (HttpWebRequest)HttpWebRequest.Create(requestUrlForWorkstepId);
                webRequest.AllowAutoRedirect = false;
                HttpWebResponse workstepIdResponse = (HttpWebResponse)webRequest.GetResponse();
                if (workstepIdResponse.StatusCode != HttpStatusCode.OK)
                {
                    Console.WriteLine("Invalid status code returned from Server: " + workstepIdResponse.StatusCode);
                    return;
                }
                Stream responseStream = workstepIdResponse.GetResponseStream();
                StreamReader reader = new StreamReader(responseStream);
                string workstepId = reader.ReadToEnd();

                // bbc) compose url for other front end
                //      see integration guide for SSP and related front ends
                //redirectionUrlFirstStep = "http://172.16.17.166:50100/Sign.aspx?WorkstepID=" + workstepId; //link to WSI on an on premise system (will only work, if workstep is created on this system too)
                redirectionUrlFirstStep = "https://beta4.testlab.xyzmo.com/SawViewer/signanywhere.aspx?WorkstepId=" + workstepId;
                #endregion

                //bc) open url
                System.Diagnostics.Process.Start(redirectionUrlFirstStep);

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
            }

        }

        static async Task CreateDraft(string[] pathToPdfFiles)
        {
            try
            {
                Envelope envObj = new Envelope(eSawAPI);
                List<string> docIds = new List<string>();

                for (int i = 0; i < pathToPdfFiles.Length; i++)
                {
                    FileStream fileToUpload = File.OpenRead(pathToPdfFiles[i]);

                    var result = await eSawAPI.SspFile.UploadTemporaryWithHttpMessagesAsync(fileToUpload, HelperGetAuthorizationHeader());
                    if (result.Response.IsSuccessStatusCode)
                    {
                        eSaw.Models.UploadSspFileResult uploadResultSuccessObj = result.Body as eSaw.Models.UploadSspFileResult;
                        docIds.Add(uploadResultSuccessObj.SspFileId);
                    }
                    else
                    {
                        eSaw.Models.Error flowError = result.Body as eSaw.Models.Error;
                        Console.WriteLine("Upload document failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                        return;
                    }
                }


                DraftCreateModel draftModel = new DraftCreateModel();
                draftModel.SspFileIds = docIds;

                //
                bool allowAgentRedirect = true;
                draftModel.CreateDraftOptions = new CreateDraftOptions("http://google.com", null, null, allowAgentRedirect, null);

                SendEnvelopeDescription envelopeDescription = new SendEnvelopeDescription();
                draftModel.SendEnvelopeDescription = envelopeDescription;
                envelopeDescription.Name = "Draft via Rest Api " + DateTime.Now.ToLongDateString() + " " + DateTime.Now.ToLongTimeString();
                envelopeDescription.EmailSubject = "my envelope via rest and draft ";
                envelopeDescription.EmailBody = "text for you ...";
                //TODO remove
                envelopeDescription.DaysUntilExpire = 28;
                
                // quite important to specify the callback to your backend system 
                // ##EnvelopeId## will be replaced by the envelopeId.
                // you can add additional parameters as well (like in the sample 'myParamForMe'
                envelopeDescription.CallbackUrl = "";

                // url to notify your back end on a status change of the envelope:
                // example:  http://172.16.17.78:57550/default.aspx?envelopeId=##EnvelopeId##action=##Action##&myParamForMe=1234</statusUpdateCallbackUrl> 
                // ##EnvelopeId## will be replaced by the envelopeId, ##Action## will be replaced by the event type that triggered the callback
                // example for status changes of an envelope: a recipient was notified, recipient signed(finished) a workstep, recipient delegated a workstep, recipient rejected a workstep
                envelopeDescription.StatusUpdateCallbackUrl = "";

                //
                envelopeDescription.Steps = new List<SendEnvelopeStep>();

                eSaw.Models.SendEnvelopeStep firstStep = new eSaw.Models.SendEnvelopeStep();
                envelopeDescription.Steps.Add(firstStep);
                firstStep.RecipientType = RecipientTypeEnum.Signer;
                firstStep.OrderIndex = 0;
                //firstStep.WorkstepConfiguration; //as it is a draft no workstep config to be set ....
                firstStep.Recipients = new List<eSaw.Models.SendEnvelopeRecipient> { new eSaw.Models.SendEnvelopeRecipient() };
                firstStep.Recipients[0].Email = "first.signer@email.com";
                firstStep.Recipients[0].LastName = "Signer";
                firstStep.Recipients[0].FirstName = "First";
                firstStep.Recipients[0].LanguageCode = "en";

                eSaw.Models.SendEnvelopeStep secondStep = new eSaw.Models.SendEnvelopeStep();
                envelopeDescription.Steps.Add(secondStep);
                //secondStep.WorkstepConfiguration; //as it is a draft no workstep config to be set ....
                secondStep.RecipientType = RecipientTypeEnum.Signer;
                secondStep.OrderIndex = 1;
                secondStep.Recipients = new List<eSaw.Models.SendEnvelopeRecipient> { new eSaw.Models.SendEnvelopeRecipient() };
                secondStep.Recipients[0].Email = "second.signer@email.com";
                secondStep.Recipients[0].LastName = "Signer";
                secondStep.Recipients[0].FirstName = "Second";
                secondStep.Recipients[0].LanguageCode = "en";

                // add copy receiver step
                eSaw.Models.SendEnvelopeStep thirdStepCopyReceiver = new eSaw.Models.SendEnvelopeStep();
                envelopeDescription.Steps.Add(thirdStepCopyReceiver);
                thirdStepCopyReceiver.RecipientType =  RecipientTypeEnum.Acknowledge;
                thirdStepCopyReceiver.OrderIndex = 2;
                thirdStepCopyReceiver.Recipients = new List<eSaw.Models.SendEnvelopeRecipient> { new eSaw.Models.SendEnvelopeRecipient() };
                thirdStepCopyReceiver.Recipients[0].Email = "ack.recipient@xyzmo.com";
                thirdStepCopyReceiver.Recipients[0].LastName = "Recipient";
                thirdStepCopyReceiver.Recipients[0].FirstName = "Ack";
                thirdStepCopyReceiver.Recipients[0].LanguageCode = "en";

                //createDraft 
                var sendEnvelopeResult = await envObj.CreateWithHttpMessagesAsync(draftModel, HelperGetAuthorizationHeader());
                string envelopeId = null;
                if (sendEnvelopeResult.Response.IsSuccessStatusCode)
                {
                    eSaw.Models.CreateDraftResult sendEnvelopeResultSuccessObj = sendEnvelopeResult.Body as CreateDraftResult;
                    envelopeId = sendEnvelopeResultSuccessObj.DraftId;
                    Console.WriteLine("Create DraftFromTemplate successful: " + envelopeId);
                }
                else
                {
                    eSaw.Models.Error flowError = sendEnvelopeResult.Body as Error;
                    Console.WriteLine("Create DraftFromTemplate failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                    return;
                }

                if (allowAgentRedirect)
                {
                    // use this link for on premise in the non  AD/LDAP version or on www.significant.com: /AgentRedirect/index?draftid=      (draft can be accessed without login)
                    // in the draftOptions, add <allowAgentRedirect>true</allowAgentRedirect>
                    System.Diagnostics.Process.Start(eSawUrl + @"/AgentRedirect/index?draftid=" + envelopeId);
                }
                else
                {
                    // use this link on www.significant.com or for on premise with AD / LDAP version: /redirect/toDefinition?Id=              (requires the user to login)
                    System.Diagnostics.Process.Start(eSawUrl + @"/redirect/toDefinition?Id=" + envelopeId);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }

        }

        static async Task SendEnvelopeFromTemplate(string templateId)
        {
            try
            {
                Envelope envObj = new Envelope(eSawAPI);

                //optional: retrieve information about template
                var templateInfoResult = await envObj.ByIdWithHttpMessagesAsync(templateId, HelperGetAuthorizationHeader());
                EnvelopeStatus templateInfo = null;
                if (templateInfoResult.Response.IsSuccessStatusCode)
                {
                    templateInfo = templateInfoResult.Body as EnvelopeStatus;

                }
                else
                {
                    eSaw.Models.Error flowError = templateInfoResult.Body as Error;
                    Console.WriteLine("Retrieving info about template failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                    return;
                }

                var copyRes = await envObj.CopyFromTemplateWithHttpMessagesAsync(templateId, HelperGetAuthorizationHeader());

                EnvelopeSendFromTemplateModel templateModel = new EnvelopeSendFromTemplateModel(templateId);

                //optional: override values for the envelope
                templateModel.EnvelopeOverrideOptions = new EnvelopeOverrideOptions();
                templateModel.EnvelopeOverrideOptions.Name = "New Envelope Name";

                //optional: override recipients, placeholder recipients must be replaced
                templateModel.EnvelopeOverrideOptions.Recipients = new List<EnvelopeOverrideOptionsRecipient>() { };

                EnvelopeOverrideOptionsRecipient newRecip1 = new EnvelopeOverrideOptionsRecipient();
                templateModel.EnvelopeOverrideOptions.Recipients.Add(newRecip1);
                // identify recipient to be replaced
                newRecip1.OrderIndex = templateInfo.Bulks[0].Steps[0].OrderIndex;
                newRecip1.Email = templateInfo.Bulks[0].Steps[0].Email;
                // assign new values for recipient
                SendEnvelopeRecipient newRecipient1 = new SendEnvelopeRecipient();
                newRecip1.Recipient = newRecipient1;
                newRecipient1.Email = "first.signer@email.com";
                newRecipient1.FirstName = "First";
                newRecipient1.LastName = "Signer";
                newRecipient1.LanguageCode = "en";

                EnvelopeOverrideOptionsRecipient newRecip2 = new EnvelopeOverrideOptionsRecipient();
                templateModel.EnvelopeOverrideOptions.Recipients.Add(newRecip2);
                // identify recipient to be replaced
                newRecip2.OrderIndex = templateInfo.Bulks[0].Steps[1].OrderIndex;
                newRecip2.Email = templateInfo.Bulks[0].Steps[1].Email;
                // assign new values for recipient
                SendEnvelopeRecipient newRecipient2 = new SendEnvelopeRecipient();
                newRecipient2.Email = "second.signer@email.com";
                newRecipient2.FirstName = "Second";
                newRecipient2.LastName = "Signer";
                newRecipient2.LanguageCode = "en";
                newRecip2.Recipient = newRecipient2;

                //send 
                var sendEnvelopeResult = await envObj.SendFromTemplateWithHttpMessagesAsync(templateModel, HelperGetAuthorizationHeader());

                if (sendEnvelopeResult.Response.IsSuccessStatusCode)
                {
                    eSaw.Models.SendEnvelopeResult sendEnvelopeResultSuccessObj = sendEnvelopeResult.Body as SendEnvelopeResult;
                    string envelopeId = sendEnvelopeResultSuccessObj.EnvelopeId;
                    Console.WriteLine("Send EnvelopeFromTemplate successful: " + envelopeId);
                }
                else
                {
                    eSaw.Models.Error flowError = sendEnvelopeResult.Body as Error;
                    Console.WriteLine("Send EnvelopeFromTemplate failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }

        static async Task CreateDraftFromTemplate(string templateId)
        {
            try
            {
                Envelope envObj = new Envelope(eSawAPI);

                //optional: retrieve information about template
                var templateInfoResult = await envObj.ByIdWithHttpMessagesAsync(templateId, HelperGetAuthorizationHeader());
                EnvelopeStatus templateInfo = null;
                if (templateInfoResult.Response.IsSuccessStatusCode)
                {
                    templateInfo = templateInfoResult.Body as EnvelopeStatus;

                }
                else
                {
                    eSaw.Models.Error flowError = templateInfoResult.Body as Error;
                    Console.WriteLine("Retrieving info about template failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                    return;
                }

                DraftCreateFromTemplateModel draftModel = new DraftCreateFromTemplateModel(templateId);

                //optional: override values for the envelope
                draftModel.EnvelopeOverrideOptions = new EnvelopeOverrideOptions();
                draftModel.EnvelopeOverrideOptions.Name = "New Envelope Name";

                //optional: override recipients, placeholder recipients must be replaced
                draftModel.EnvelopeOverrideOptions.Recipients = new List<EnvelopeOverrideOptionsRecipient>() { };

                EnvelopeOverrideOptionsRecipient newRecip1 = new EnvelopeOverrideOptionsRecipient();
                draftModel.EnvelopeOverrideOptions.Recipients.Add(newRecip1);
                // identify recipient to be replaced
                newRecip1.OrderIndex = templateInfo.Bulks[0].Steps[0].OrderIndex;
                newRecip1.Email = templateInfo.Bulks[0].Steps[0].Email;
                // assign new values for recipient
                SendEnvelopeRecipient newRecipient1 = new SendEnvelopeRecipient();
                newRecip1.Recipient = newRecipient1;
                newRecipient1.Email = "first.signer@email.com";
                newRecipient1.FirstName = "First";
                newRecipient1.LastName = "Signer";
                newRecipient1.LanguageCode = "en";

                EnvelopeOverrideOptionsRecipient newRecip2 = new EnvelopeOverrideOptionsRecipient();
                draftModel.EnvelopeOverrideOptions.Recipients.Add(newRecip2);
                // identify recipient to be replaced
                newRecip2.OrderIndex = templateInfo.Bulks[0].Steps[1].OrderIndex;
                newRecip2.Email = templateInfo.Bulks[0].Steps[1].Email;
                // assign new values for recipient
                SendEnvelopeRecipient newRecipient2 = new SendEnvelopeRecipient();
                newRecipient2.Email = "second.signer@email.com";
                newRecipient2.FirstName = "Second";
                newRecipient2.LastName = "Signer";
                newRecipient2.LanguageCode = "en";
                newRecip2.Recipient = newRecipient2;
                
                //
                bool allowAgentRedirect = true;
                draftModel.CreateDraftOptions = new CreateDraftOptions("http://google.com", null, null, allowAgentRedirect, null);

                //createDraft 
                var sendEnvelopeResult = await envObj.CreateFromTemplateWithHttpMessagesAsync(draftModel, HelperGetAuthorizationHeader());
                string envelopeId = null;
                if (sendEnvelopeResult.Response.IsSuccessStatusCode)
                {
                    eSaw.Models.CreateDraftResult sendEnvelopeResultSuccessObj = sendEnvelopeResult.Body as CreateDraftResult;
                    envelopeId = sendEnvelopeResultSuccessObj.DraftId;
                    Console.WriteLine("Create DraftFromTemplate successful: " + envelopeId);
                }
                else
                {
                    eSaw.Models.Error flowError = sendEnvelopeResult.Body as Error;
                    Console.WriteLine("Create DraftFromTemplate failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                    return;
                }

                if (allowAgentRedirect)
                {
                    // use this link for on premise in the non  AD/LDAP version or on www.significant.com: /AgentRedirect/index?draftid=      (draft can be accessed without login)
                    // in the draftOptions, add <allowAgentRedirect>true</allowAgentRedirect>
                    System.Diagnostics.Process.Start(eSawUrl + @"/AgentRedirect/index?draftid=" + envelopeId);
                }
                else
                {
                    // use this link on www.significant.com or for on premise with AD / LDAP version: /redirect/toDefinition?Id=              (requires the user to login)
                    System.Diagnostics.Process.Start(eSawUrl + @"/redirect/toDefinition?Id=" + envelopeId);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }

        static async Task GetEnvelopeStatus(string envelopeId)
        {
            try
            {
                var result = await eSawAPI.Envelope.ByIdWithHttpMessagesAsync(envelopeId, HelperGetAuthorizationHeader());

                if (result.Response.IsSuccessStatusCode)
                {
                    eSaw.Models.EnvelopeStatus envelopeStatusObj = result.Body as EnvelopeStatus;
                    Console.WriteLine("Status of Envelope with " + envelopeId + " is: " + envelopeStatusObj.Status);
                }
                else
                {
                    eSaw.Models.Error flowError = result.Body as Error;
                    Console.WriteLine("Get Envelope returned: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }

        static async Task DownloadCompletedEnvelope(string envelopeId)
        {
            try
            {
                var result = await eSawAPI.Envelope.ByIdWithHttpMessagesAsync(envelopeId, HelperGetAuthorizationHeader());

                if (result.Response.IsSuccessStatusCode)
                {
                    eSaw.Models.EnvelopeStatus envelopeStatusObj = result.Body as EnvelopeStatus;

                    if (envelopeStatusObj.Status !=  EnvelopeStatusEnum.Completed &&
                        envelopeStatusObj.Status !=  EnvelopeStatusEnum.BulkPartlyCompleted &&
                        envelopeStatusObj.Status != EnvelopeStatusEnum.BulkCompleted)
                    {
                        Console.WriteLine("Envelope is not completed, status is: " + envelopeStatusObj.Status);
                        return;
                    }
                    else
                    {
                        DirectoryInfo di = System.IO.Directory.CreateDirectory(@"..\..\..\output\" + envelopeId);
                        Console.WriteLine("Writing files to " + di.FullName);
                        DirectoryInfo diForSubEnv = di;

                        for (int i = 0; i < envelopeStatusObj.Bulks.Count; i++)
                        {
                            eSaw.Models.EnvelopeBulkStatus subEnvelope = envelopeStatusObj.Bulks[i];

                            if (envelopeStatusObj.Bulks.Count > 1)
                            {
                                diForSubEnv = new DirectoryInfo(di.FullName + "\\" + subEnvelope.Email.Replace("@", "(at)"));
                                diForSubEnv.Create();
                            }

                            if (subEnvelope.Status ==   EnvelopeStatusEnum.Completed)
                            {
                                for (int f = 0; f < subEnvelope.FinishedDocuments.Count; f++)
                                {
                                    string docName = subEnvelope.FinishedDocuments[f].FileName;
                                    Console.WriteLine(docName);
                                    if (subEnvelope.FinishedDocuments[f].FormFields == null || subEnvelope.FinishedDocuments[f].FormFields.Count <= 0)
                                    {
                                        Console.WriteLine("No form fields for this document!");
                                    }
                                    else
                                    {
                                        for (int ff = 0; ff < subEnvelope.FinishedDocuments[f].FormFields.Count; ff++)
                                        {
                                            string msg = "\t" + subEnvelope.FinishedDocuments[f].FormFields[ff].Name + ": ";
                                            if (subEnvelope.FinishedDocuments[f].FormFields[ff].Values.Count != 0)
                                            {
                                                msg += subEnvelope.FinishedDocuments[f].FormFields[ff].Values[0].Value;

                                            }
                                            Console.WriteLine(msg);
                                        }
                                    }

                                    var documentDownloadRes = await eSawAPI.Envelope.DownloadCompletedDocumentWithHttpMessagesAsync(subEnvelope.FinishedDocuments[f].FlowDocumentId, HelperGetAuthorizationHeader());
                                    if (documentDownloadRes.Response.IsSuccessStatusCode)
                                    {
                                        string filePath = diForSubEnv.FullName + @"\" + subEnvelope.FinishedDocuments[f].FileName;
                                        byte[] fileContent = (documentDownloadRes.Body as MemoryStream).ToArray();
                                        File.WriteAllBytes(filePath, fileContent);
                                        Console.WriteLine("Document downloaded to: " + filePath);
                                    }
                                    else
                                    {
                                        eSaw.Models.Error downloadError = documentDownloadRes.Body as Error;
                                        Console.WriteLine("Error on downloading document: " + downloadError.Message + ", ErrorId: " + downloadError.ErrorId);
                                    }

                                    
                                } //end document loop

                                //download envelopeLog
                                var envelopeLogDownloadRes = await eSawAPI.Envelope.DownloadCompletedDocumentWithHttpMessagesAsync(subEnvelope.LogDocumentId, HelperGetAuthorizationHeader());
                                if (envelopeLogDownloadRes.Response.IsSuccessStatusCode)
                                {
                                    string filePath = diForSubEnv.FullName + @"\" + "EnvelopeLog.pdf";
                                    byte[] fileContent = (envelopeLogDownloadRes.Body as MemoryStream).ToArray();
                                    File.WriteAllBytes(filePath, fileContent);
                                    Console.WriteLine("EnvelopeLog downloaded to: " + filePath);
                                }
                                else
                                {
                                    eSaw.Models.Error downloadError = envelopeLogDownloadRes.Body as Error;
                                    Console.WriteLine("Error on downloading document: " + downloadError.Message + ", ErrorId: " + downloadError.ErrorId);
                                }
                            }
                            else
                            {
                                Console.WriteLine("\tSubEnvelope for recipient " + subEnvelope.Email + " is in status: " + subEnvelope.Status);
                            }
                        }
                    }
                }
                else
                {
                    eSaw.Models.Error flowError = result.Body as Error;
                    Console.WriteLine("Get Envelope returned: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }
        static async Task DeleteEnvelope(string envelopeId)
        {
            try
            {
                var result = await eSawAPI.Envelope.DeleteWithHttpMessagesAsync(envelopeId, HelperGetAuthorizationHeader());
                if (result.Response.StatusCode == HttpStatusCode.NoContent)
                {
                    Console.WriteLine("Delete envelope successful");
                }

                else
                {
                    eSaw.Models.Error flowError = result.Body as eSaw.Models.Error;
                    Console.WriteLine("Delete envelope failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }

        }

        static async Task FindEnvelopes()
        {
            try
            {
                FindEnvelopesDescriptor envDescriptor = new FindEnvelopesDescriptor();
                envDescriptor.Status = FindEnvelopeStatusEnum.Completed;

                var result = await eSawAPI.Envelope.FindWithHttpMessagesAsync(envDescriptor, HelperGetAuthorizationHeader());
                if (result.Response.StatusCode == HttpStatusCode.OK)
                {
                    eSaw.Models.ExtendedFindEnvelopesResult envelopeFindResObj = result.Body as ExtendedFindEnvelopesResult;
                    if (envelopeFindResObj.Envelopes != null && envelopeFindResObj.Envelopes.Count > 0)
                    {
                        Console.WriteLine("Find envelopes returns " + envelopeFindResObj.Envelopes.Count + " envelopes!");
                        for (int i = 0; i < envelopeFindResObj.Envelopes.Count; i++)
                        {
                            Console.WriteLine("\t" + envelopeFindResObj.Envelopes[i].Id + " " + envelopeFindResObj.Envelopes[i].Status);
                        }
                    }
                    else
                    {
                        Console.WriteLine("No envelopes found on FindEnvelopes!");
                    }
                }

                else
                {
                    eSaw.Models.Error flowError = result.Body as eSaw.Models.Error;
                    Console.WriteLine("Find envelopes failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                }

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }

        static async Task FindAllEnvelopsThatAreWaitingForDedicatedRecipientAndReplaceRecipient()
        {
            // this is useful, in case a signer is not available and someone else should sign instead of this signer (employee is on vacation, vacation replacement should sign instead)

            try
            {

                string recipientEmailToLookUp = "first.signer@email.com";
                SendEnvelopeRecipient newRecipient = new SendEnvelopeRecipient();
                newRecipient.FirstName = "New";
                newRecipient.LastName = "Recipient";
                newRecipient.Email = "new.recipient@email.com";
                newRecipient.LanguageCode = "en";
                ReplaceRecipientData replaceRecipData = new ReplaceRecipientData(newRecipient, null);

                FindEnvelopesDescriptor envDescriptor = new FindEnvelopesDescriptor();                
                envDescriptor.WaitingForRecipient = recipientEmailToLookUp;

                var result = await eSawAPI.Envelope.FindWithHttpMessagesAsync(envDescriptor, HelperGetAuthorizationHeader());
                eSaw.Models.ExtendedFindEnvelopesResult envelopeFindResObj = null;
                if (result.Response.StatusCode == HttpStatusCode.OK)
                {                
                    envelopeFindResObj = result.Body as ExtendedFindEnvelopesResult;
                    if (envelopeFindResObj.Envelopes == null && envelopeFindResObj.Envelopes.Count == 0)
                    {
                        Console.WriteLine("No envelopes found on FindEnvelopes!");
                        return;
                    }                    
                }
                else
                {
                    eSaw.Models.Error flowError = result.Body as eSaw.Models.Error;
                    Console.WriteLine("Find envelopes failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                    return;
                }

                Console.WriteLine("Find envelopes returned " + envelopeFindResObj.Envelopes.Count + " envelopes that are waiting for "+recipientEmailToLookUp+"!");
                for (int i = 0; i < envelopeFindResObj.Envelopes.Count; i++)
                {
                    string envelopeId = envelopeFindResObj.Envelopes[i].Id;

                    var envStatusResult = await eSawAPI.Envelope.ByIdWithHttpMessagesAsync(envelopeId, HelperGetAuthorizationHeader());

                    if (!envStatusResult.Response.IsSuccessStatusCode)
                    {
                        eSaw.Models.Error flowError = result.Body as Error;
                        Console.WriteLine("Get Envelope status for envelopeId " + envelopeId + " returned: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                        continue;
                    }

                    eSaw.Models.EnvelopeStatus envelopeStatusObj = envStatusResult.Body as EnvelopeStatus;
                    Console.WriteLine("Status of Envelope with " + envelopeId + " is: " + envelopeStatusObj.Status);

                    string recipientId = null;
                    for (   int s = 0; s < envelopeStatusObj.Bulks[0].Steps.Count; s++) {
                        if (envelopeStatusObj.Bulks[0].Steps[s].Email == recipientEmailToLookUp &&
                            (envelopeStatusObj.Bulks[0].Steps[s].RecipientType ==  RecipientTypeEnum.Signer || envelopeStatusObj.Bulks[0].Steps[s].RecipientType ==  RecipientTypeEnum.Pkcs7Signer) &&
                            envelopeStatusObj.Bulks[0].Steps[s].Status ==   WorkstepResultEnum.NotSigned)
                        {
                            recipientId = envelopeStatusObj.Bulks[0].Steps[s].Id;
                            break;
                        }
                    }                      

                    var recipReplaceResult = await eSawAPI.Recipient.ReplaceWithHttpMessagesAsync(envelopeId, recipientId, replaceRecipData, HelperGetAuthorizationHeader());
                    if (recipReplaceResult.Response.StatusCode == HttpStatusCode.NoContent)
                    {
                        Console.WriteLine("Replace recipient for envelope with id " + envelopeId + " was successful!");
                    }
                    else {
                        eSaw.Models.Error flowError = recipReplaceResult.Body as eSaw.Models.Error;
                        Console.WriteLine("Replace recipient failed for envelope with id " + envelopeId + " : " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                    }

                }

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }

        static async Task FindUsers()
        {
            try
            {
                FindUsersDescriptor findUserDesc = new FindUsersDescriptor();
                

                var result = await eSawAPI.User.FindWithHttpMessagesAsync(findUserDesc, HelperGetAuthorizationHeader());
                if (result.Response.StatusCode == HttpStatusCode.OK)
                {
                    eSaw.Models.ExtendedFindUsersResult userFindRes = result.Body as ExtendedFindUsersResult;

                    if (userFindRes.Entries != null && userFindRes.Entries.Count > 0)
                    {
                        Console.WriteLine("Find users returns " + userFindRes.Entries.Count + " users!");
                        for (int i = 0; i < userFindRes.Entries.Count; i++)
                        {
                            Console.WriteLine("\t" + userFindRes.Entries[i].Email + " " + userFindRes.Entries[i].Id);
                        }
                    }
                    else
                    {
                        Console.WriteLine("No envelopes found on FindEnvelopes!");
                    }
                }

                else
                {
                    eSaw.Models.Error flowError = result.Body as eSaw.Models.Error;
                    Console.WriteLine("Find users failed: " + flowError.Message + ", ErrorId: " + flowError.ErrorId);
                }

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }

    }
}

